# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic query."""

import functools
from typing import Any, Callable, Type, Union

import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.structured import mapping
from langfun.core.structured import schema as schema_lib
import pyglove as pg


@lf.use_init_args(['schema', 'default', 'examples'])
class QueryStructure(mapping.Mapping):
  """Query an object out from a natural language text."""

  context_title = 'CONTEXT'
  input_title = 'INPUT_OBJECT'

  # Mark schema as required.
  schema: pg.typing.Annotated[
      schema_lib.schema_spec(), 'Required schema for parsing.'
  ]


class QueryStructureJson(QueryStructure):
  """Query a structured value using JSON as the protocol."""

  preamble = """
      Please respond to the last {{ input_title }} with {{ output_title}} according to {{ schema_title }}:

      INSTRUCTIONS:
        1. If the schema has `_type`, carry it over to the JSON output.
        2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

      {{ input_title }}:
        1 + 1 =

      {{ schema_title }}:
        {"result": {"_type": "langfun.core.structured.prompting.Answer", "final_answer": int}}

      {{ output_title}}:
        {"result": {"_type": "langfun.core.structured.prompting.Answer", "final_answer": 2}}
      """

  protocol = 'json'
  schema_title = 'SCHEMA'
  output_title = 'JSON'


class QueryStructurePython(QueryStructure):
  """Query a structured value using Python as the protocol."""

  preamble = """
      Please respond to the last {{ input_title }} with {{ output_title }} according to {{ schema_title }}.

      {{ input_title }}:
        1 + 1 =

      {{ schema_title }}:
        Answer

        ```python
        class Answer:
          final_answer: int
        ```

      {{ output_title }}:
        ```python
        Answer(
          final_answer=2
        )
        ```
      """
  protocol = 'python'
  schema_title = 'OUTPUT_TYPE'
  output_title = 'OUTPUT_OBJECT'


def _query_structure_cls(
    protocol: schema_lib.SchemaProtocol,
) -> Type[QueryStructure]:
  if protocol == 'json':
    return QueryStructureJson
  elif protocol == 'python':
    return QueryStructurePython
  else:
    raise ValueError(f'Unknown protocol: {protocol!r}.')


def query(
    prompt: Union[str, pg.Symbolic],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ] = None,
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    cache_seed: int | None = 0,
    response_postprocess: Callable[[str], str] | None = None,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    returns_message: bool = False,
    skip_lm: bool = False,
    **kwargs,
) -> Any:
  """Parse a natural langugage message based on schema.

  Examples:

    ```
    class FlightDuration:
      hours: int
      minutes: int

    class Flight(pg.Object):
      airline: str
      flight_number: str
      departure_airport_code: str
      arrival_airport_code: str
      departure_time: str
      arrival_time: str
      duration: FlightDuration
      stops: int
      price: float

    prompt = '''
      Information about flight UA2631.
      '''

    r = lf.query(prompt, Flight)
    assert isinstance(r, Flight)
    assert r.airline == 'United Airlines'
    assert r.departure_airport_code == 'SFO'
    assert r.duration.hour = 7
    ```

  Args:
    prompt: A str (may contain {{}} as template) as natural language input, or a
      `pg.Symbolic` object as structured input as prompt to LLM.
    schema: A type annotation as the schema for output object. If str (default),
      the response will be a str in natural language.
    default: The default value if parsing failed. If not specified, error will
      be raised.
    lm: The language model to use. If not specified, the language model from
      `lf.context` context manager will be used.
    examples: An optional list of fewshot examples for helping parsing. If None,
      the default one-shot example will be added.
    cache_seed: Seed for computing cache key. The cache key is determined by a
      tuple of (lm, prompt, cache seed). If None, cache will be disabled for the
      query even cache is configured by the LM.
    response_postprocess: An optional callable object to process the raw LM
      response before parsing it into the final output object. If None, the raw
      LM response will not be processed.
    autofix: Number of attempts to auto fix the generated code. If 0, autofix is
      disabled. Auto-fix is not supported for 'json' protocol.
    autofix_lm: The language model to use for autofix. If not specified, the
      `autofix_lm` from `lf.context` context manager will be used. Otherwise it
      will use `lm`.
    protocol: The protocol for schema/value representation. Applicable values
      are 'json' and 'python'. By default `python` will be used.
    returns_message: If True, returns `lf.Message` as the output, instead of
      returning the structured `message.result`.
    skip_lm: If True, returns the rendered prompt as a UserMessage object.
      otherwise return the LLM response based on the rendered prompt.
    **kwargs: Keyword arguments passed to render the prompt or configure the 
      `lf.structured.Mapping` class. Notable kwargs are:
      - template_str: Change the root template for query.
      - preamble: Change the preamble for query.
      - mapping_template: Change the template for each mapping examle.

  Returns:
    The result based on the schema.
  """
    # Internal usage logging.

  # When `lf.query` is used for symbolic completion, schema is automatically
  # inferred when it is None.
  if isinstance(prompt, pg.Symbolic) and prompt.sym_partial and schema is None:
    schema = prompt.__class__

  # Create a copy of the prompt if it has a parent object, so all child modality
  # objects could be referred by path relative to the prompt.
  if isinstance(prompt, lf.Template) and prompt.sym_parent:
    prompt = prompt.clone()

  if schema in (None, str):
    # Query with natural language output.
    output = lf.LangFunc.from_value(prompt, **kwargs)(
        lm=lm, cache_seed=cache_seed, skip_lm=skip_lm
    )
    if response_postprocess:
      processed_text = response_postprocess(output.text)
      if processed_text != output.text:
        output = lf.AIMessage(processed_text, source=output)
    return output if returns_message else output.text

  # Query with structured output.
  prompt_kwargs = kwargs.copy()

  # NOTE(daiyip): when `template_str` is passed in, it's intended to modify the
  # QueryStructure template string. Therefore, we pop out the argument for
  # prompt rendering.
  prompt_kwargs.pop('template_str', None)

  if isinstance(prompt, (str, lf.Message, lf.Template)):
    prompt = lf.Template.from_value(prompt, **prompt_kwargs).render(lm=lm)
  else:
    prompt = schema_lib.mark_missing(prompt)

  output = _query_structure_cls(protocol)(
      input=prompt,
      schema=schema,
      default=default,
      examples=examples,
      response_postprocess=response_postprocess,
      autofix=autofix if protocol == 'python' else 0,
      **kwargs,
  )(
      lm=lm,
      autofix_lm=autofix_lm or lm,
      cache_seed=cache_seed,
      skip_lm=skip_lm,
  )
  return output if returns_message else output.result


def query_prompt(
    prompt: Union[str, pg.Symbolic],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ] = None,
    **kwargs,
) -> lf.Message:
  """Returns the final prompt sent to LLM for `lf.query`."""
  kwargs.pop('returns_message', None)
  kwargs.pop('skip_lm', None)
  return query(prompt, schema, skip_lm=True, returns_message=True, **kwargs)


def query_output(
    response: Union[str, lf.Message],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ],
    **kwargs,
) -> Any:
  """Returns the final output of `lf.query` from a provided LLM response."""
  kwargs.pop('prompt', None)
  kwargs.pop('lm', None)
  return query(
      'Unused prompt', schema, lm=fake.StaticResponse(response), **kwargs
  )


def query_reward(
    mapping_example: Union[str, mapping.MappingExample],
    response: Union[str, lf.Message],
) -> float | None:
  """Returns the reward of an LLM response based on an mapping example."""
  if isinstance(mapping_example, str):
    mapping_example = pg.from_json_str(mapping_example)
    assert isinstance(mapping_example, mapping.MappingExample), mapping_example
  schema = mapping_example.schema

  if schema and isinstance(schema.spec, pg.typing.Object):
    output_cls = schema.spec.cls
  elif schema is None and isinstance(mapping_example.output, pg.Object):
    output_cls = mapping_example.output.__class__
  else:
    output_cls = None

  reward_fn = _reward_fn(output_cls)
  if reward_fn is None:
    return None

  return reward_fn(
      query_output(response, output_cls),
      mapping_example.input,
      mapping_example.output,
      mapping_example.metadata,
  )


@functools.cache
def _reward_fn(cls) -> Callable[
    [
        pg.Object,    # Actual output object.
        Any,          # Input object.
        pg.Object,    # Expected output object.
        pg.Dict       # User metadata.
    ], float] | None:
  """Returns the reward function for a class that is being queried."""
  if not callable(getattr(cls, '__reward__', None)):
    return None

  signature = pg.typing.signature(cls.__reward__)
  num_args = len(signature.args)
  if num_args < 2 or num_args > 4:
    raise TypeError(
        f'`{cls.__type_name__}.__reward__` should have signature: '
        '`__reward__(self, input, [expected_output], [expected_metadata])`.'
    )
  def _reward(self, input, expected_output, metadata):  # pylint: disable=redefined-builtin
    args = [self, input, expected_output, metadata]
    return cls.__reward__(*args[:num_args])
  return _reward



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reporting evaluation results."""

import time
from typing import Annotated

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib

Runner = experiment_lib.Runner
Experiment = experiment_lib.Experiment
Example = example_lib.Example


_SUMMARY_FILE = 'summary.html'
_EVALULATION_DETAIL_FILE = 'index.html'


class HtmlReporter(experiment_lib.Plugin):
  """Plugin for periodically generating HTML reports for the experiment."""

  summary_interval: Annotated[
      int,
      'The interval of writing summary in seconds.'
  ] = 60

  experiment_report_interval: Annotated[
      int,
      'The interval of writing report for inidividual experiments in seconds.'
  ] = 60

  def _on_bound(self):
    super()._on_bound()
    self._last_summary_time = 0
    self._last_experiment_report_time = {}

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    self._maybe_update_summary(runner)
    self._last_experiment_report_time = {leaf.id: 0 for leaf in root.leaf_nodes}

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    self._maybe_update_summary(runner, force=True)

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    if experiment.is_leaf:
      self._maybe_update_experiment_html(runner, experiment)

  def on_experiment_complete(
      self, runner: Runner, experiment: Experiment
  ):
    if experiment.is_leaf:
      self._maybe_update_experiment_html(runner, experiment, force=True)

  def on_example_complete(
      self, runner: Runner, experiment: Experiment, example: Example
  ):
    self._save_example_html(runner, experiment, example)
    self._maybe_update_experiment_html(runner, experiment)
    self._maybe_update_summary(runner)

  def _maybe_update_summary(self, runner: Runner, force: bool = False) -> None:
    """Maybe update the summary of current run."""
    run = runner.current_run
    def _summary():
      run.experiment.to_html(
          collapse_level=None,
          extra_flags=dict(
              current_run=run, interactive=False, card_view=True,
          )
      ).save(
          run.output_path_for(run.experiment, _SUMMARY_FILE)
      )

    if force or (time.time() - self._last_summary_time > self.summary_interval):
      runner.background_run(_summary)
      self._last_summary_time = time.time()

  def _maybe_update_experiment_html(
      self, runner: Runner, experiment: Experiment, force: bool = False
  ) -> None:
    def _save():
      html = experiment.to_html(
          collapse_level=None,
          extra_flags=dict(
              current_run=runner.current_run,
              interactive=False,
              card_view=False,
          ),
      )
      html.save(
          runner.current_run.output_path_for(
              experiment, _EVALULATION_DETAIL_FILE
          )
      )
    if force or (
        time.time() - self._last_experiment_report_time[experiment.id]
        > self.experiment_report_interval
    ):
      runner.background_run(_save)
      self._last_experiment_report_time[experiment.id] = time.time()

  def _save_example_html(
      self, runner: Runner, experiment: Experiment, example: Example
  ) -> None:
    """Saves the example."""
    def _save():
      html = example.to_html(
          collapse_level=None,
          enable_summary_tooltip=False,
          extra_flags=dict(
              # For properly rendering the next link.
              num_examples=getattr(experiment, 'num_examples', None)
          ),
      )
      html.save(
          runner.current_run.output_path_for(
              experiment, f'{example.id}.html'
          )
      )
    runner.background_run(_save)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checkpointing evaluation runs."""
import threading

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
import pyglove as pg

Example = example_lib.Example
Experiment = experiment_lib.Experiment
Runner = experiment_lib.Runner


class Checkpointer(experiment_lib.Plugin):
  """Plugin for checkpointing evaluation runs."""

  checkpoint_filename: str = 'checkpoint.bagz'

  def _on_bound(self):
    super()._on_bound()
    self._lock = threading.Lock()
    self._state_writer = None

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment,
  ) -> None:
    self._state_writer = {}

  def on_run_abort(
      self,
      runner: Runner,
      root: Experiment,
      error: BaseException
  ) -> None:
    with self._lock:
      if self._state_writer is not None:
        self._state_writer.clear()

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment,
  ) -> None:
    with self._lock:
      assert self._state_writer is not None and not self._state_writer

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    """Creates the checkpoint file."""
    if not experiment.is_leaf:
      return
    # For refresh runs, we don't want to load the previous state.
    if not runner.current_run.refresh:
      experiment.load_state(
          runner.current_run.input_path_for(
              experiment, self.checkpoint_filename
          ),
          raise_if_not_exist=False
      )
    state_writer = StateWriter(
        runner.current_run.output_path_for(
            experiment, self.checkpoint_filename
        )
    )
    with self._lock:
      if self._state_writer is not None:
        self._state_writer[experiment.id] = state_writer

  def on_experiment_complete(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    """Closes the checkpoint file."""
    if not experiment.is_leaf:
      return
    assert experiment.id in self._state_writer
    with self._lock:
      if self._state_writer is not None:
        del self._state_writer[experiment.id]

  def on_example_complete(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    assert experiment.id in self._state_writer
    if not example.has_error:
      runner.background_run(self._state_writer[experiment.id].add, example)


class StateWriter:
  """Thread safe state writer."""

  def __init__(self, path: str):
    self._lock = threading.Lock()
    self._sequence_writer = pg.io.open_sequence(path, 'w')

  def add(self, example: Example):
    example_blob = pg.to_json_str(
        example,
        hide_default_values=True,
        save_ref_value=True,
        exclude_input=True
    )
    with self._lock:
      if self._sequence_writer is None:
        return
      self._sequence_writer.add(example_blob)

  def __del__(self):
    # Make sure there is no write in progress.
    with self._lock:
      assert self._sequence_writer is not None
      self._sequence_writer.close()
      self._sequence_writer = None



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
import unittest

from langfun.core.eval.v2 import reporting
from langfun.core.eval.v2 import runners as runners_lib  # pylint: disable=unused-import
from langfun.core.eval.v2 import test_helper
import pyglove as pg


class ReportingTest(unittest.TestCase):

  def test_reporting(self):
    root_dir = os.path.join(tempfile.gettempdir(), 'test_reporting')
    experiment = test_helper.test_experiment()
    reporter = reporting.HtmlReporter()
    run = experiment.run(root_dir, 'new', plugins=[reporter])
    pg.io.path_exists(run.output_path_for(experiment, 'summary.html'))
    for leaf in experiment.leaf_nodes:
      self.assertTrue(
          pg.io.path_exists(run.output_path_for(leaf, 'index.html'))
      )
      for i in range(leaf.num_examples):
        pg.io.path_exists(run.output_path_for(leaf, f'{i + 1}.html'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Gemini models."""

import os
import unittest
from unittest import mock

from google.cloud.aiplatform import models as aiplatform_models
from vertexai import generative_models
import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import vertexai
import pyglove as pg


example_image = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


def mock_generate_content(content, generation_config, **kwargs):
  del kwargs
  c = pg.Dict(generation_config.to_dict())
  return generative_models.GenerationResponse.from_dict({
      'candidates': [
          {
              'index': 0,
              'content': {
                  'role': 'model',
                  'parts': [
                      {
                          'text': (
                              f'This is a response to {content[0]} with '
                              f'temperature={c.temperature}, '
                              f'top_p={c.top_p}, '
                              f'top_k={c.top_k}, '
                              f'max_tokens={c.max_output_tokens}, '
                              f'stop={"".join(c.stop_sequences)}.'
                          )
                      },
                  ],
              },
          },
      ]
  })


def mock_endpoint_predict(instances, **kwargs):
  del kwargs
  assert len(instances) == 1
  return aiplatform_models.Prediction(
      predictions=[
          f"This is a response to {instances[0]['prompt']} with"
          f" temperature={instances[0]['temperature']},"
          f" top_p={instances[0]['top_p']}, top_k={instances[0]['top_k']},"
          f" max_tokens={instances[0]['max_tokens']}."
      ],
      deployed_model_id='',
  )


class VertexAITest(unittest.TestCase):
  """Tests for Vertex model."""

  def test_content_from_message_text_only(self):
    text = 'This is a beautiful day'
    model = vertexai.VertexAIGeminiPro1()
    chunks = model._content_from_message(lf.UserMessage(text))
    self.assertEqual(chunks, [text])

  def test_content_from_message_mm(self):
    message = lf.UserMessage(
        'This is an <<[[image]]>>, what is it?',
        image=lf_modalities.Image.from_bytes(example_image),
    )

    # Non-multimodal model.
    with self.assertRaisesRegex(lf.ModalityError, 'Unsupported modality'):
      vertexai.VertexAIGeminiPro1()._content_from_message(message)

    model = vertexai.VertexAIGeminiPro1Vision()
    chunks = model._content_from_message(message)
    self.maxDiff = None
    self.assertEqual([chunks[0], chunks[2]], ['This is an', ', what is it?'])
    self.assertIsInstance(chunks[1], generative_models.Part)

  def test_generation_response_to_message_text_only(self):
    response = generative_models.GenerationResponse.from_dict({
        'candidates': [
            {
                'index': 0,
                'content': {
                    'role': 'model',
                    'parts': [
                        {
                            'text': 'hello world',
                        },
                    ],
                },
            },
        ],
    })
    model = vertexai.VertexAIGeminiPro1()
    message = model._generation_response_to_message(response)
    self.assertEqual(message, lf.AIMessage('hello world'))

  def test_model_hub(self):
    with mock.patch(
        'vertexai.generative_models.'
        'GenerativeModel.__init__'
    ) as mock_model_init:
      mock_model_init.side_effect = lambda *args, **kwargs: None
      model = vertexai._VERTEXAI_MODEL_HUB.get_generative_model(
          'gemini-1.0-pro'
      )
      self.assertIsNotNone(model)
      self.assertIs(
          vertexai._VERTEXAI_MODEL_HUB.get_generative_model('gemini-1.0-pro'),
          model,
      )

    with mock.patch(
        'vertexai.language_models.'
        'TextGenerationModel.from_pretrained'
    ) as mock_model_init:

      class TextGenerationModel:
        pass

      mock_model_init.side_effect = lambda *args, **kw: TextGenerationModel()
      model = vertexai._VERTEXAI_MODEL_HUB.get_text_generation_model(
          'text-bison'
      )
      self.assertIsNotNone(model)
      self.assertIs(
          vertexai._VERTEXAI_MODEL_HUB.get_text_generation_model('text-bison'),
          model,
      )

  def test_project_and_location_check(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `project`'):
      _ = vertexai.VertexAIGeminiPro1()._api_initialized

    with self.assertRaisesRegex(ValueError, 'Please specify `location`'):
      _ = vertexai.VertexAIGeminiPro1(project='abc')._api_initialized

    self.assertTrue(
        vertexai.VertexAIGeminiPro1(
            project='abc', location='us-central1'
        )._api_initialized
    )

    os.environ['VERTEXAI_PROJECT'] = 'abc'
    os.environ['VERTEXAI_LOCATION'] = 'us-central1'
    self.assertTrue(vertexai.VertexAIGeminiPro1()._api_initialized)
    del os.environ['VERTEXAI_PROJECT']
    del os.environ['VERTEXAI_LOCATION']

  def test_generation_config(self):
    model = vertexai.VertexAIGeminiPro1()
    json_schema = {
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
        },
        'required': ['name'],
        'title': 'Person',
    }
    config = model._generation_config(
        lf.UserMessage('hi', json_schema=json_schema),
        lf.LMSamplingOptions(
            temperature=2.0,
            top_p=1.0,
            top_k=20,
            max_tokens=1024,
            stop=['\n'],
        ),
    )
    actual = config.to_dict()
    # There is a discrepancy between the `property_ordering` in the
    # Google-internal version and the open-source version.
    actual['response_schema'].pop('property_ordering', None)
    if pg.KeyPath.parse('response_schema.type_').get(actual):
      actual['response_schema']['type'] = actual['response_schema'].pop('type_')
    if pg.KeyPath.parse('response_schema.properties.name.type_').get(actual):
      actual['response_schema']['properties']['name']['type'] = actual[
          'response_schema']['properties']['name'].pop('type_')

    self.assertEqual(
        actual,
        dict(
            temperature=2.0,
            top_p=1.0,
            top_k=20.0,
            max_output_tokens=1024,
            stop_sequences=['\n'],
            response_mime_type='application/json',
            response_schema={
                'type': 'OBJECT',
                'properties': {
                    'name': {'type': 'STRING'}
                },
                'required': ['name'],
                'title': 'Person',
            }
        ),
    )
    with self.assertRaisesRegex(
        ValueError, '`json_schema` must be a dict, got'
    ):
      model._generation_config(
          lf.UserMessage('hi', json_schema='not a dict'),
          lf.LMSamplingOptions(),
      )

  def test_call_generative_model(self):
    with mock.patch(
        'vertexai.generative_models.'
        'GenerativeModel.__init__'
    ) as mock_model_init:
      mock_model_init.side_effect = lambda *args, **kwargs: None

      with mock.patch(
          'vertexai.generative_models.'
          'GenerativeModel.generate_content'
      ) as mock_generate:
        mock_generate.side_effect = mock_generate_content

        lm = vertexai.VertexAIGeminiPro1(project='abc', location='us-central1')
        self.assertEqual(
            lm(
                'hello',
                temperature=2.0,
                top_p=1.0,
                top_k=20,
                max_tokens=1024,
                stop='\n',
            ).text,
            (
                'This is a response to hello with temperature=2.0, '
                'top_p=1.0, top_k=20.0, max_tokens=1024, stop=\n.'
            ),
        )

  def test_call_text_generation_model(self):
    with mock.patch(
        'vertexai.language_models.'
        'TextGenerationModel.from_pretrained'
    ) as mock_model_init:

      class TextGenerationModel:

        def predict(self, prompt, **kwargs):
          c = pg.Dict(kwargs)
          return pg.Dict(
              text=(
                  f'This is a response to {prompt} with '
                  f'temperature={c.temperature}, '
                  f'top_p={c.top_p}, '
                  f'top_k={c.top_k}, '
                  f'max_tokens={c.max_output_tokens}, '
                  f'stop={"".join(c.stop_sequences)}.'
              )
          )

      mock_model_init.side_effect = lambda *args, **kw: TextGenerationModel()
      lm = vertexai.VertexAIPalm2(project='abc', location='us-central1')
      self.assertEqual(
          lm(
              'hello',
              temperature=2.0,
              top_p=1.0,
              top_k=20,
              max_tokens=1024,
              stop='\n',
          ).text,
          (
              'This is a response to hello with temperature=2.0, '
              'top_p=1.0, top_k=20, max_tokens=1024, stop=\n.'
          ),
      )

  def test_call_endpoint_model(self):
    with mock.patch(
        'google.cloud.aiplatform.models.Endpoint.__init__'
    ) as mock_model_init:
      mock_model_init.side_effect = lambda *args, **kwargs: None
      with mock.patch(
          'google.cloud.aiplatform.models.Endpoint.predict'
      ) as mock_model_predict:

        mock_model_predict.side_effect = mock_endpoint_predict
        lm = vertexai.VertexAI(
            'custom',
            endpoint_name='123',
            project='abc',
            location='us-central1',
        )
        self.assertEqual(
            lm(
                'hello',
                temperature=2.0,
                top_p=1.0,
                top_k=20,
                max_tokens=50,
            ),
            'This is a response to hello with temperature=2.0, top_p=1.0,'
            ' top_k=20, max_tokens=50.',
        )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun LLM implementations."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top

# LMs for testing.
from langfun.core.llms.fake import Fake
from langfun.core.llms.fake import Echo
from langfun.core.llms.fake import StaticMapping
from langfun.core.llms.fake import StaticResponse
from langfun.core.llms.fake import StaticSequence

# Compositional models.
from langfun.core.llms.compositional import RandomChoice

# REST-based models.
from langfun.core.llms.rest import REST

# Gemini models.
from langfun.core.llms.google_genai import GenAI
from langfun.core.llms.google_genai import GeminiExp_20241114
from langfun.core.llms.google_genai import GeminiFlash1_5
from langfun.core.llms.google_genai import GeminiPro
from langfun.core.llms.google_genai import GeminiPro1_5
from langfun.core.llms.google_genai import GeminiProVision
from langfun.core.llms.google_genai import Palm2
from langfun.core.llms.google_genai import Palm2_IT

# OpenAI models.
from langfun.core.llms.openai import OpenAI

from langfun.core.llms.openai import GptO1Preview
from langfun.core.llms.openai import GptO1Preview_20240912
from langfun.core.llms.openai import GptO1Mini
from langfun.core.llms.openai import GptO1Mini_20240912

from langfun.core.llms.openai import Gpt4oMini
from langfun.core.llms.openai import Gpt4oMini_20240718
from langfun.core.llms.openai import Gpt4o
from langfun.core.llms.openai import Gpt4o_20241120
from langfun.core.llms.openai import Gpt4o_20240806
from langfun.core.llms.openai import Gpt4o_20240513

from langfun.core.llms.openai import Gpt4Turbo
from langfun.core.llms.openai import Gpt4Turbo_20240409
from langfun.core.llms.openai import Gpt4TurboPreview
from langfun.core.llms.openai import Gpt4TurboPreview_20240125
from langfun.core.llms.openai import Gpt4TurboPreview_20231106
from langfun.core.llms.openai import Gpt4VisionPreview
from langfun.core.llms.openai import Gpt4VisionPreview_20231106
from langfun.core.llms.openai import Gpt4
from langfun.core.llms.openai import Gpt4_20230613

from langfun.core.llms.openai import Gpt4_32K
from langfun.core.llms.openai import Gpt4_32K_20230613

from langfun.core.llms.openai import Gpt35Turbo
from langfun.core.llms.openai import Gpt35Turbo_20240125
from langfun.core.llms.openai import Gpt35Turbo_20231106
from langfun.core.llms.openai import Gpt35Turbo_20230613
from langfun.core.llms.openai import Gpt35Turbo16K
from langfun.core.llms.openai import Gpt35Turbo16K_20230613

# For backward compatibility.
Gpt4TurboPreview_0125 = Gpt4TurboPreview_20240125
Gpt4TurboPreview_1106 = Gpt4TurboPreview_20231106
Gpt4VisionPreview_1106 = Gpt4VisionPreview_20231106
Gpt4_0613 = Gpt4_20230613
Gpt4_32K_0613 = Gpt4_32K_20230613
Gpt35Turbo_0125 = Gpt35Turbo_20240125
Gpt35Turbo_1106 = Gpt35Turbo_20231106
Gpt35Turbo_0613 = Gpt35Turbo_20230613
Gpt35Turbo16K_0613 = Gpt35Turbo16K_20230613

from langfun.core.llms.openai import Gpt35

from langfun.core.llms.openai import Gpt3
from langfun.core.llms.openai import Gpt3Curie
from langfun.core.llms.openai import Gpt3Babbage
from langfun.core.llms.openai import Gpt3Ada

from langfun.core.llms.anthropic import Anthropic
from langfun.core.llms.anthropic import Claude35Sonnet
from langfun.core.llms.anthropic import Claude35Sonnet20241022
from langfun.core.llms.anthropic import Claude35Sonnet20240620
from langfun.core.llms.anthropic import Claude3Opus
from langfun.core.llms.anthropic import Claude3Sonnet
from langfun.core.llms.anthropic import Claude3Haiku
from langfun.core.llms.anthropic import VertexAIAnthropic
from langfun.core.llms.anthropic import VertexAIClaude3_5_Sonnet_20241022
from langfun.core.llms.anthropic import VertexAIClaude3_5_Sonnet_20240620
from langfun.core.llms.anthropic import VertexAIClaude3_5_Haiku_20241022

from langfun.core.llms.groq import Groq
from langfun.core.llms.groq import GroqLlama3_2_3B
from langfun.core.llms.groq import GroqLlama3_2_1B
from langfun.core.llms.groq import GroqLlama3_1_70B
from langfun.core.llms.groq import GroqLlama3_1_8B
from langfun.core.llms.groq import GroqLlama3_70B
from langfun.core.llms.groq import GroqLlama3_8B
from langfun.core.llms.groq import GroqLlama2_70B
from langfun.core.llms.groq import GroqMistral_8x7B
from langfun.core.llms.groq import GroqGemma2_9B_IT
from langfun.core.llms.groq import GroqGemma_7B_IT
from langfun.core.llms.groq import GroqWhisper_Large_v3
from langfun.core.llms.groq import GroqWhisper_Large_v3Turbo

from langfun.core.llms.vertexai import VertexAI
from langfun.core.llms.vertexai import VertexAIGemini1_5
from langfun.core.llms.vertexai import VertexAIGeminiPro1_5
from langfun.core.llms.vertexai import VertexAIGeminiPro1_5_Latest
from langfun.core.llms.vertexai import VertexAIGeminiPro1_5_001
from langfun.core.llms.vertexai import VertexAIGeminiPro1_5_002
from langfun.core.llms.vertexai import VertexAIGeminiPro1_5_0514
from langfun.core.llms.vertexai import VertexAIGeminiPro1_5_0409
from langfun.core.llms.vertexai import VertexAIGeminiFlash1_5_Latest
from langfun.core.llms.vertexai import VertexAIGeminiFlash1_5
from langfun.core.llms.vertexai import VertexAIGeminiFlash1_5_001
from langfun.core.llms.vertexai import VertexAIGeminiFlash1_5_002
from langfun.core.llms.vertexai import VertexAIGeminiFlash1_5_0514
from langfun.core.llms.vertexai import VertexAIGeminiPro1
from langfun.core.llms.vertexai import VertexAIGeminiPro1Vision
from langfun.core.llms.vertexai import VertexAIPalm2
from langfun.core.llms.vertexai import VertexAIPalm2_32K
from langfun.core.llms.vertexai import VertexAICustom


# LLaMA C++ models.
from langfun.core.llms.llama_cpp import LlamaCppRemote

# Placeholder for Google-internal imports.

# Include cache as sub-module.
from langfun.core.llms import cache

# pylint: enable=g-import-not-at-top
# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for base action."""

import unittest

import langfun.core as lf
from langfun.core.agentic import action as action_lib


class SessionTest(unittest.TestCase):

  def test_basics(self):
    test = self

    class Bar(action_lib.Action):

      def call(self, session, **kwargs):
        test.assertIs(session.current_invocation.action, self)
        session.info('Begin Bar')
        return 2

    class Foo(action_lib.Action):
      x: int

      def call(self, session, **kwargs):
        test.assertIs(session.current_invocation.action, self)
        session.info('Begin Foo', x=1)
        return self.x + Bar()(session)

    session = action_lib.Session()
    root = session.root_invocation
    self.assertIsInstance(root.action, action_lib.RootAction)
    self.assertIs(session.current_invocation, session.root_invocation)
    self.assertEqual(Foo(1)(session), 3)
    self.assertEqual(len(session.root_invocation.child_invocations), 1)
    self.assertEqual(len(session.root_invocation.child_invocations[0].logs), 1)
    self.assertEqual(
        len(session.root_invocation.child_invocations[0].child_invocations),
        1
    )
    self.assertEqual(
        len(session.root_invocation
            .child_invocations[0].child_invocations[0].logs),
        1
    )
    self.assertEqual(
        len(session.root_invocation
            .child_invocations[0].child_invocations[0].child_invocations),
        0
    )
    self.assertIs(session.current_invocation, session.root_invocation)
    self.assertIs(session.final_result, 3)
    self.assertIn(
        'invocation-final-result',
        session.to_html().content,
    )

  def test_log(self):
    session = action_lib.Session()
    session.debug('hi', x=1, y=2)
    session.info('hi', x=1, y=2)
    session.warning('hi', x=1, y=2)
    session.error('hi', x=1, y=2)
    session.fatal('hi', x=1, y=2)

  def test_as_message(self):
    session = action_lib.Session()
    self.assertIsInstance(session.as_message(), lf.AIMessage)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vertex AI generative models."""

import functools
import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
import pyglove as pg

try:
  # pylint: disable=g-import-not-at-top
  from google.auth import credentials as credentials_lib
  import vertexai
  from google.cloud.aiplatform import models as aiplatform_models
  from vertexai import generative_models
  from vertexai import language_models
  # pylint: enable=g-import-not-at-top

  Credentials = credentials_lib.Credentials
except ImportError:
  credentials_lib = None   # pylint: disable=invalid-name
  vertexai = None
  generative_models = None
  language_models = None
  aiplatform_models = None
  Credentials = Any


# https://cloud.google.com/vertex-ai/generative-ai/pricing
# describes that the average number of characters per token is about 4.
AVGERAGE_CHARS_PER_TOEKN = 4


# Price in US dollars,
# from https://cloud.google.com/vertex-ai/generative-ai/pricing
# as of 2024-10-10.
SUPPORTED_MODELS_AND_SETTINGS = {
    'gemini-1.5-pro-001': pg.Dict(
        api='gemini',
        rpm=100,
        cost_per_1k_input_chars=0.0003125,
        cost_per_1k_output_chars=0.00125,
    ),
    'gemini-1.5-pro-002': pg.Dict(
        api='gemini',
        rpm=100,
        cost_per_1k_input_chars=0.0003125,
        cost_per_1k_output_chars=0.00125,
    ),
    'gemini-1.5-flash-002': pg.Dict(
        api='gemini',
        rpm=500,
        cost_per_1k_input_chars=0.00001875,
        cost_per_1k_output_chars=0.000075,
    ),
    'gemini-1.5-flash-001': pg.Dict(
        api='gemini',
        rpm=500,
        cost_per_1k_input_chars=0.00001875,
        cost_per_1k_output_chars=0.000075,
    ),
    'gemini-1.5-pro': pg.Dict(
        api='gemini',
        rpm=100,
        cost_per_1k_input_chars=0.0003125,
        cost_per_1k_output_chars=0.00125,
    ),
    'gemini-1.5-flash': pg.Dict(
        api='gemini',
        rpm=500,
        cost_per_1k_input_chars=0.00001875,
        cost_per_1k_output_chars=0.000075,
    ),
    'gemini-1.5-pro-latest': pg.Dict(
        api='gemini',
        rpm=100,
        cost_per_1k_input_chars=0.0003125,
        cost_per_1k_output_chars=0.00125,
    ),
    'gemini-1.5-flash-latest': pg.Dict(
        api='gemini',
        rpm=500,
        cost_per_1k_input_chars=0.00001875,
        cost_per_1k_output_chars=0.000075,
    ),
    'gemini-1.5-pro-preview-0514': pg.Dict(
        api='gemini',
        rpm=50,
        cost_per_1k_input_chars=0.0003125,
        cost_per_1k_output_chars=0.00125,
    ),
    'gemini-1.5-pro-preview-0409': pg.Dict(
        api='gemini',
        rpm=50,
        cost_per_1k_input_chars=0.0003125,
        cost_per_1k_output_chars=0.00125,
    ),
    'gemini-1.5-flash-preview-0514': pg.Dict(
        api='gemini',
        rpm=200,
        cost_per_1k_input_chars=0.00001875,
        cost_per_1k_output_chars=0.000075,
    ),
    'gemini-1.0-pro': pg.Dict(
        api='gemini',
        rpm=300,
        cost_per_1k_input_chars=0.000125,
        cost_per_1k_output_chars=0.000375,
    ),
    'gemini-1.0-pro-vision': pg.Dict(
        api='gemini',
        rpm=100,
        cost_per_1k_input_chars=0.000125,
        cost_per_1k_output_chars=0.000375,
    ),
    # PaLM APIs.
    'text-bison': pg.Dict(
        api='palm',
        rpm=1600
    ),
    'text-bison-32k': pg.Dict(
        api='palm',
        rpm=300
    ),
    'text-unicorn': pg.Dict(
        api='palm',
        rpm=100
    ),
    # Endpoint
    # TODO(chengrun): Set a more appropriate rpm for endpoint.
    'custom': pg.Dict(api='endpoint', rpm=20),
}


@lf.use_init_args(['model'])
class VertexAI(lf.LanguageModel):
  """Language model served on VertexAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      (
          'Vertex AI model name. See '
          'https://cloud.google.com/vertex-ai/generative-ai/docs/learn/models '
          'for details.'
      ),
  ]

  endpoint_name: pg.typing.Annotated[
      str | None,
      'Vertex Endpoint name or ID.',
  ]

  project: Annotated[
      str | None,
      (
          'Vertex AI project ID. Or set from environment variable '
          'VERTEXAI_PROJECT.'
      ),
  ] = None

  location: Annotated[
      str | None,
      (
          'Vertex AI service location. Or set from environment variable '
          'VERTEXAI_LOCATION.'
      ),
  ] = None

  credentials: Annotated[
      Credentials | None,
      (
          'Credentials to use. If None, the default credentials to the '
          'environment will be used.'
      ),
  ] = None

  supported_modalities: Annotated[
      list[str],
      'A list of MIME types for supported modalities'
  ] = []

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('_api_initialized', None)
    if generative_models is None:
      raise RuntimeError(
          'Please install "langfun[llm-google-vertex]" to use Vertex AI models.'
      )

  @functools.cached_property
  def _api_initialized(self):
    project = self.project or os.environ.get('VERTEXAI_PROJECT', None)
    if not project:
      raise ValueError(
          'Please specify `project` during `__init__` or set environment '
          'variable `VERTEXAI_PROJECT` with your Vertex AI project ID.'
      )

    location = self.location or os.environ.get('VERTEXAI_LOCATION', None)
    if not location:
      raise ValueError(
          'Please specify `location` during `__init__` or set environment '
          'variable `VERTEXAI_LOCATION` with your Vertex AI service location.'
      )

    credentials = self.credentials
    # Placeholder for Google-internal credentials.
    assert vertexai is not None
    vertexai.init(project=project, location=location, credentials=credentials)
    return True

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'VertexAI({self.model})'

  @property
  def resource_id(self) -> str:
    """Returns a string to identify the resource for rate control."""
    return self.model_id

  @property
  def max_concurrency(self) -> int:
    """Returns the maximum number of concurrent requests."""
    return self.rate_to_max_concurrency(
        requests_per_min=SUPPORTED_MODELS_AND_SETTINGS[self.model].rpm,
        tokens_per_min=0,
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1k_input_chars = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_input_chars', None
    )
    cost_per_1k_output_chars = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_output_chars', None
    )
    if cost_per_1k_output_chars is None or cost_per_1k_input_chars is None:
      return None
    return (
        cost_per_1k_input_chars * num_input_tokens
        + cost_per_1k_output_chars * num_output_tokens
    ) * AVGERAGE_CHARS_PER_TOEKN / 1000

  def _generation_config(
      self, prompt: lf.Message, options: lf.LMSamplingOptions
  ) -> Any:  # generative_models.GenerationConfig
    """Creates generation config from langfun sampling options."""
    assert generative_models is not None
    # Users could use `metadata_json_schema` to pass additional
    # request arguments.
    json_schema = prompt.metadata.get('json_schema')
    response_mime_type = None
    if json_schema is not None:
      if not isinstance(json_schema, dict):
        raise ValueError(
            f'`json_schema` must be a dict, got {json_schema!r}.'
        )
      response_mime_type = 'application/json'
      prompt.metadata.formatted_text = (
          prompt.text
          + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
          + pg.to_json_str(json_schema, json_indent=2)
      )

    return generative_models.GenerationConfig(
        temperature=options.temperature,
        top_p=options.top_p,
        top_k=options.top_k,
        max_output_tokens=options.max_tokens,
        stop_sequences=options.stop,
        response_mime_type=response_mime_type,
        response_schema=json_schema,
    )

  def _content_from_message(
      self, prompt: lf.Message
  ) -> list[str | Any]:
    """Gets generation input from langfun message."""
    assert generative_models is not None
    chunks = []

    for lf_chunk in prompt.chunk():
      if isinstance(lf_chunk, str):
        chunks.append(lf_chunk)
      elif isinstance(lf_chunk, lf_modalities.Mime):
        try:
          modalities = lf_chunk.make_compatible(
              self.supported_modalities + ['text/plain']
          )
          if isinstance(modalities, lf_modalities.Mime):
            modalities = [modalities]
          for modality in modalities:
            if modality.is_text:
              chunk = modality.to_text()
            else:
              chunk = generative_models.Part.from_data(
                  modality.to_bytes(), modality.mime_type
              )
            chunks.append(chunk)
        except lf.ModalityError as e:
          raise lf.ModalityError(f'Unsupported modality: {lf_chunk!r}') from e
      else:
        raise lf.ModalityError(f'Unsupported modality: {lf_chunk!r}')
    return chunks

  def _generation_response_to_message(
      self,
      response: Any,  # generative_models.GenerationResponse
  ) -> lf.Message:
    """Parses generative response into message."""
    return lf.AIMessage(response.text)

  def _generation_endpoint_response_to_message(
      self,
      response: Any,  # google.cloud.aiplatform.aiplatform.models.Prediction
  ) -> lf.Message:
    """Parses Endpoint response into message."""
    return lf.AIMessage(response.predictions[0])

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    assert self._api_initialized, 'Vertex AI API is not initialized.'
    # TODO(yifenglu): It seems this exception is due to the instability of the
    # API. We should revisit this later.
    retry_on_errors = [
        (Exception, 'InternalServerError'),
        (Exception, 'ResourceExhausted'),
        (Exception, '_InactiveRpcError'),
        (Exception, 'ValueError'),
    ]

    return self._parallel_execute_with_currency_control(
        self._sample_single,
        prompts,
        retry_on_errors=retry_on_errors,
    )

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    if self.sampling_options.n > 1:
      raise ValueError(
          f'`n` greater than 1 is not supported: {self.sampling_options.n}.'
      )
    api = SUPPORTED_MODELS_AND_SETTINGS[self.model].api
    match api:
      case 'gemini':
        return self._sample_generative_model(prompt)
      case 'palm':
        return self._sample_text_generation_model(prompt)
      case 'endpoint':
        return self._sample_endpoint_model(prompt)
      case _:
        raise ValueError(f'Unsupported API: {api}')

  def _sample_generative_model(self, prompt: lf.Message) -> lf.LMSamplingResult:
    """Samples a generative model."""
    model = _VERTEXAI_MODEL_HUB.get_generative_model(self.model)
    input_content = self._content_from_message(prompt)
    response = model.generate_content(
        input_content,
        generation_config=self._generation_config(
            prompt, self.sampling_options
        ),
    )
    usage_metadata = response.usage_metadata
    usage = lf.LMSamplingUsage(
        prompt_tokens=usage_metadata.prompt_token_count,
        completion_tokens=usage_metadata.candidates_token_count,
        total_tokens=usage_metadata.total_token_count,
        estimated_cost=self.estimate_cost(
            num_input_tokens=usage_metadata.prompt_token_count,
            num_output_tokens=usage_metadata.candidates_token_count,
        ),
    )
    return lf.LMSamplingResult(
        [
            # Scoring is not supported.
            lf.LMSample(
                self._generation_response_to_message(response), score=0.0
            ),
        ],
        usage=usage,
    )

  def _sample_text_generation_model(
      self, prompt: lf.Message
  ) -> lf.LMSamplingResult:
    """Samples a text generation model."""
    model = _VERTEXAI_MODEL_HUB.get_text_generation_model(self.model)
    predict_options = dict(
        temperature=self.sampling_options.temperature,
        top_k=self.sampling_options.top_k,
        top_p=self.sampling_options.top_p,
        max_output_tokens=self.sampling_options.max_tokens,
        stop_sequences=self.sampling_options.stop,
    )
    response = model.predict(prompt.text, **predict_options)
    return lf.LMSamplingResult([
        # Scoring is not supported.
        lf.LMSample(lf.AIMessage(response.text), score=0.0)
    ])

  def _sample_endpoint_model(self, prompt: lf.Message) -> lf.LMSamplingResult:
    """Samples a text generation model."""
    assert aiplatform_models is not None
    model = aiplatform_models.Endpoint(self.endpoint_name)
    # TODO(chengrun): Add support for stop_sequences.
    predict_options = dict(
        temperature=self.sampling_options.temperature
        if self.sampling_options.temperature is not None
        else 1.0,
        top_k=self.sampling_options.top_k
        if self.sampling_options.top_k is not None
        else 32,
        top_p=self.sampling_options.top_p
        if self.sampling_options.top_p is not None
        else 1,
        max_tokens=self.sampling_options.max_tokens
        if self.sampling_options.max_tokens is not None
        else 8192,
    )
    instances = [{'prompt': prompt.text, **predict_options}]
    response = model.predict(instances=instances)

    return lf.LMSamplingResult([
        # Scoring is not supported.
        lf.LMSample(
            self._generation_endpoint_response_to_message(response), score=0.0
        )
    ])


class _ModelHub:
  """Vertex AI model hub."""

  def __init__(self):
    self._generative_model_cache = {}
    self._text_generation_model_cache = {}

  def get_generative_model(
      self, model_id: str
  ) -> Any:  # generative_models.GenerativeModel:
    """Gets a generative model by model id."""
    model = self._generative_model_cache.get(model_id, None)
    if model is None:
      assert generative_models is not None
      model = generative_models.GenerativeModel(model_id)
      self._generative_model_cache[model_id] = model
    return model

  def get_text_generation_model(
      self, model_id: str
  ) -> Any:  # language_models.TextGenerationModel
    """Gets a text generation model by model id."""
    model = self._text_generation_model_cache.get(model_id, None)
    if model is None:
      assert language_models is not None
      model = language_models.TextGenerationModel.from_pretrained(model_id)
      self._text_generation_model_cache[model_id] = model
    return model


_VERTEXAI_MODEL_HUB = _ModelHub()


_IMAGE_TYPES = [
    'image/png',
    'image/jpeg',
    'image/webp',
    'image/heic',
    'image/heif',
]

_AUDIO_TYPES = [
    'audio/aac',
    'audio/flac',
    'audio/mp3',
    'audio/m4a',
    'audio/mpeg',
    'audio/mpga',
    'audio/mp4',
    'audio/opus',
    'audio/pcm',
    'audio/wav',
    'audio/webm'
]

_VIDEO_TYPES = [
    'video/mov',
    'video/mpeg',
    'video/mpegps',
    'video/mpg',
    'video/mp4',
    'video/webm',
    'video/wmv',
    'video/x-flv',
    'video/3gpp',
]

_DOCUMENT_TYPES = [
    'application/pdf',
    'text/plain',
    'text/csv',
    'text/html',
    'text/xml',
    'text/x-script.python',
    'application/json',
]


class VertexAIGemini1_5(VertexAI):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 model."""

  supported_modalities: pg.typing.List(str).freeze(   # pytype: disable=invalid-annotation
      _DOCUMENT_TYPES + _IMAGE_TYPES + _AUDIO_TYPES + _VIDEO_TYPES
  )


class VertexAIGeminiPro1_5_Latest(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro-latest'


class VertexAIGeminiPro1_5(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro'


class VertexAIGeminiPro1_5_002(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro-002'


class VertexAIGeminiPro1_5_001(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro-001'


class VertexAIGeminiPro1_5_0514(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro preview model."""

  model = 'gemini-1.5-pro-preview-0514'


class VertexAIGeminiPro1_5_0409(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro preview model."""

  model = 'gemini-1.5-pro-preview-0409'


class VertexAIGeminiFlash1_5_Latest(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash-latest'


class VertexAIGeminiFlash1_5(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash'


class VertexAIGeminiFlash1_5_002(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash-002'


class VertexAIGeminiFlash1_5_001(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash-001'


class VertexAIGeminiFlash1_5_0514(VertexAIGemini1_5):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash preview model."""

  model = 'gemini-1.5-flash-preview-0514'


class VertexAIGeminiPro1(VertexAI):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.0 Pro model."""

  model = 'gemini-1.0-pro'


class VertexAIGeminiPro1Vision(VertexAI):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.0 Pro model."""

  model = 'gemini-1.0-pro-vision'
  supported_modalities: pg.typing.List(str).freeze(   # pytype: disable=invalid-annotation
      _IMAGE_TYPES + _VIDEO_TYPES
  )


class VertexAIPalm2(VertexAI):  # pylint: disable=invalid-name
  """Vertex AI PaLM2 text generation model."""

  model = 'text-bison'


class VertexAIPalm2_32K(VertexAI):  # pylint: disable=invalid-name
  """Vertex AI PaLM2 text generation model (32K context length)."""

  model = 'text-bison-32k'


class VertexAICustom(VertexAI):  # pylint: disable=invalid-name
  """Vertex AI Custom model (Endpoint)."""

  model = 'custom'



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.repr_utils."""

import io
import unittest

from langfun.core import repr_utils
import pyglove as pg


class SharingContentTest(unittest.TestCase):

  def test_sharing(self):
    s = io.StringIO()

    self.assertTrue(repr_utils.write_maybe_shared(s, '<hr>'))
    self.assertTrue(repr_utils.write_maybe_shared(s, '<hr>'))

    with repr_utils.share_parts() as ctx1:
      self.assertTrue(repr_utils.write_maybe_shared(s, '<style></style>'))
      self.assertFalse(repr_utils.write_maybe_shared(s, '<style></style>'))

      with repr_utils.share_parts() as ctx2:
        self.assertIs(ctx2, ctx1)
        self.assertFalse(repr_utils.write_maybe_shared(s, '<style></style>'))
        self.assertTrue(repr_utils.write_maybe_shared(s, '<style>a</style>'))
        self.assertFalse(repr_utils.write_maybe_shared(s, '<style>a</style>'))
        self.assertTrue(repr_utils.write_maybe_shared(s, '<style>b</style>'))

      with repr_utils.share_parts() as ctx3:
        self.assertIs(ctx3, ctx1)
        self.assertFalse(repr_utils.write_maybe_shared(s, '<style></style>'))
        self.assertFalse(repr_utils.write_maybe_shared(s, '<style>a</style>'))
        self.assertFalse(repr_utils.write_maybe_shared(s, '<style>a</style>'))
        self.assertFalse(repr_utils.write_maybe_shared(s, '<style>b</style>'))

    self.assertEqual(
        s.getvalue(),
        '<hr><hr><style></style><style>a</style><style>b</style>'
    )
    self.assertEqual(ctx1['<style></style>'], 4)
    self.assertEqual(ctx1['<style>b</style>'], 2)
    self.assertEqual(ctx1['<style>a</style>'], 4)

  def test_escape_quoted(self):
    self.assertEqual(
        repr_utils.escape_quoted(str('<a>')), '<a>'
    )
    self.assertEqual(
        repr_utils.escape_quoted('x=<a>, b="<a>"'),
        'x=<a>, b="&lt;a&gt;"'
    )

  def test_html(self):
    html = repr_utils.Html('<div>foo</div>')
    self.assertEqual(html.content, '<div>foo</div>')
    self.assertEqual(html._repr_html_(), '<div>foo</div>')

  def test_html_repr(self):
    class Foo(pg.Object):
      x: int

    class Bar(pg.Object):

      def _repr_html_(self):
        return '<bar>'

    html = repr_utils.html_repr(
        {'foo': pg.Ref(Foo(1)), 'bar': Bar(), 'baz': '<lf_image>'}
    )
    self.assertIn('foo</span>', html)
    self.assertIn('<bar>', html)
    self.assertIn('&lt;lf_image&gt;', html)
    self.assertNotIn('Ref', html)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Anthropic models."""

import base64
import os
from typing import Any
import unittest
from unittest import mock

from google.auth import exceptions
from langfun.core import language_model
from langfun.core import message as lf_message
from langfun.core import modalities as lf_modalities
from langfun.core.llms import anthropic
import pyglove as pg
import requests


def mock_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs

  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'content': [{
          'type': 'text',
          'text': (
              f'hello with temperature={json.get("temperature")}, '
              f'top_k={json.get("top_k")}, '
              f'top_p={json.get("top_p")}, '
              f'max_tokens={json.get("max_tokens")}, '
              f'stop={json.get("stop_sequences")}.'
          ),
      }],
      'usage': {
          'input_tokens': 2,
          'output_tokens': 1,
      },
  }).encode()
  return response


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)

pdf_content = (
    b'%PDF-1.4\n1 0 obj\n<< /Type /Catalog /Pages 2 0 R >>\nendobj\n2 0 obj\n<<'
    b' /Type /Pages /Count 1 /Kids [3 0 R] >>\nendobj\n3 0 obj\n<< /Type /Page'
    b' /Parent 2 0 R /MediaBox [0 0 612 792] /Contents 4 0 R >>\nendobj\n4 0'
    b' obj\n<< /Length 44 >>\nstream\nBT /F1 24 Tf 100 700 Td (Hello, PDF'
    b' content!) Tj ET\nendstream\nendobj\n5 0 obj\n<< /Type /Font /Subtype'
    b' /Type1 /BaseFont /Helvetica >>\nendobj\nxref\n0 6\n0000000000 65535 f'
    b' \n0000000010 00000 n \n0000000079 00000 n \n0000000178 00000 n'
    b' \n0000000278 00000 n \n0000000407 00000 n \ntrailer\n<< /Size 6 /Root 1'
    b' 0 R >>\nstartxref\n517\n%%EOF'
)


def mock_mm_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs
  v = json['messages'][0]['content'][0]
  content = lf_modalities.Mime.from_bytes(base64.b64decode(v['source']['data']))

  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'content': [{
          'type': 'text',
          'text': f'{v["type"]}: {content.mime_type}',
      }],
      'usage': {
          'input_tokens': 2,
          'output_tokens': 1,
      },
  }).encode()
  return response


def mock_requests_post_error(status_code, error_type, error_message):
  def _mock_requests(url: str, json: dict[str, Any], **kwargs):
    del url, json, kwargs
    response = requests.Response()
    response.status_code = status_code
    response._content = pg.to_json_str(
        {
            'error': {
                'type': error_type,
                'message': error_message,
            }
        }
    ).encode()
    return response

  return _mock_requests


class AnthropicTest(unittest.TestCase):

  def test_basics(self):
    self.assertEqual(
        anthropic.Claude3Haiku().model_id, 'claude-3-haiku-20240307'
    )
    self.assertGreater(anthropic.Claude3Haiku().max_concurrency, 0)

  def test_api_key(self):
    lm = anthropic.Claude3Haiku()
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      lm('hi')

    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post

      lm = anthropic.Claude3Haiku(api_key='fake key')
      self.assertRegex(lm('hi').text, 'hello.*')

      os.environ['ANTHROPIC_API_KEY'] = 'abc'
      lm = anthropic.Claude3Haiku()
      self.assertRegex(lm('hi').text, 'hello.*')
      del os.environ['ANTHROPIC_API_KEY']

  def test_call(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      response = lm('hello', temperature=0.0, top_k=0.1, top_p=0.2, stop=['\n'])
      self.assertEqual(
          response.text,
          (
              'hello with temperature=0.0, top_k=0.1, top_p=0.2, '
              "max_tokens=4096, stop=['\\n']."
          ),
      )
      self.assertIsNotNone(response.usage)
      self.assertIsNotNone(response.usage.prompt_tokens, 2)
      self.assertIsNotNone(response.usage.completion_tokens, 1)
      self.assertIsNotNone(response.usage.total_tokens, 3)

  def test_mm_call(self):
    with mock.patch('requests.Session.post') as mock_mm_request:
      mock_mm_request.side_effect = mock_mm_requests_post
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      response = lm(lf_modalities.Image.from_bytes(image_content), lm=lm)
      self.assertEqual(response.text, 'image: image/png')

  def test_pdf_call(self):
    with mock.patch('requests.Session.post') as mock_mm_request:
      mock_mm_request.side_effect = mock_mm_requests_post
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      response = lm(lf_modalities.PDF.from_bytes(pdf_content), lm=lm)
      self.assertEqual(response.text, 'document: application/pdf')

  def test_call_errors(self):
    for status_code, error_type, error_message in [
        (429, 'rate_limit', 'Rate limit exceeded.'),
        (529, 'service_unavailable', 'Service unavailable.'),
        (500, 'bad_request', 'Bad request.'),
    ]:
      with mock.patch('requests.Session.post') as mock_mm_request:
        mock_mm_request.side_effect = mock_requests_post_error(
            status_code, error_type, error_message
        )
        lm = anthropic.Claude3Haiku(api_key='fake_key')
        with self.assertRaisesRegex(
            Exception, f'.*{status_code}: .*{error_message}'
        ):
          lm('hello', max_attempts=1)


class VertexAIAnthropicTest(unittest.TestCase):
  """Tests for VertexAI Anthropic models."""

  def test_basics(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `project`'):
      lm = anthropic.VertexAIClaude3_5_Sonnet_20241022()
      lm('hi')

    model = anthropic.VertexAIClaude3_5_Sonnet_20241022(project='langfun')

    # NOTE(daiyip): For OSS users, default credentials are not available unless
    # users have already set up their GCP project. Therefore we ignore the
    # exception here.
    try:
      model._initialize()
    except exceptions.DefaultCredentialsError:
      pass

    self.assertEqual(
        model.api_endpoint,
        (
            'https://us-east5-aiplatform.googleapis.com/v1/projects/'
            'langfun/locations/us-east5/publishers/anthropic/'
            'models/claude-3-5-sonnet-v2@20241022:streamRawPredict'
        )
    )
    request = model.request(
        lf_message.UserMessage('hi'),
        language_model.LMSamplingOptions(temperature=0.0),
    )
    self.assertEqual(
        request,
        {
            'anthropic_version': 'vertex-2023-10-16',
            'max_tokens': 8192,
            'messages': [
                {'content': [{'text': 'hi', 'type': 'text'}], 'role': 'user'}
            ],
            'stream': False,
            'temperature': 0.0,
            'top_k': 40,
        },
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from OpenAI."""

import collections
import functools
import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
import pyglove as pg

try:
  import openai  # pylint: disable=g-import-not-at-top

  if hasattr(openai, 'error'):
    # For lower versions.
    ServiceUnavailableError = openai.error.ServiceUnavailableError
    RateLimitError = openai.error.RateLimitError
    APITimeoutError = (
        openai.error.APIError,
        '.*The server had an error processing your request'
    )
  else:
    # For higher versions.
    ServiceUnavailableError = getattr(openai, 'InternalServerError')
    RateLimitError = getattr(openai, 'RateLimitError')
    APITimeoutError = getattr(openai, 'APITimeoutError')
except ImportError:
  openai = None


# From https://platform.openai.com/settings/organization/limits
_DEFAULT_TPM = 250000
_DEFAULT_RPM = 3000

SUPPORTED_MODELS_AND_SETTINGS = {
    # Models from https://platform.openai.com/docs/models
    # RPM is from https://platform.openai.com/docs/guides/rate-limits
    # o1 (preview) models.
    # Pricing in US dollars, from https://openai.com/api/pricing/
    # as of 2024-10-10.
    'o1-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.015,
        cost_per_1k_output_tokens=0.06,
    ),
    'o1-preview-2024-09-12': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.015,
        cost_per_1k_output_tokens=0.06,
    ),
    'o1-mini': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.012,
    ),
    'o1-mini-2024-09-12': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.012,
    ),
    # GPT-4o models
    'gpt-4o-mini': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.00015,
        cost_per_1k_output_tokens=0.0006,
    ),
    'gpt-4o-mini-2024-07-18': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.00015,
        cost_per_1k_output_tokens=0.0006,
    ),
    'gpt-4o': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0025,
        cost_per_1k_output_tokens=0.01,
    ),
    'gpt-4o-2024-11-20': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0025,
        cost_per_1k_output_tokens=0.01,
    ),
    'gpt-4o-2024-08-06': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0025,
        cost_per_1k_output_tokens=0.01,
    ),
    'gpt-4o-2024-05-13': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.005,
        cost_per_1k_output_tokens=0.015,
    ),
    # GPT-4-Turbo models
    'gpt-4-turbo': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-turbo-2024-04-09': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-turbo-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-0125-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-1106-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-vision-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-1106-vision-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    # GPT-4 models
    'gpt-4': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.03,
        cost_per_1k_output_tokens=0.06,
    ),
    'gpt-4-0613': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.03,
        cost_per_1k_output_tokens=0.06,
    ),
    'gpt-4-0314': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.03,
        cost_per_1k_output_tokens=0.06,
    ),
    'gpt-4-32k': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.06,
        cost_per_1k_output_tokens=0.12,
    ),
    'gpt-4-32k-0613': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.06,
        cost_per_1k_output_tokens=0.12,
    ),
    'gpt-4-32k-0314': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.06,
        cost_per_1k_output_tokens=0.12,
    ),
    # GPT-3.5-Turbo models
    'gpt-3.5-turbo': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0005,
        cost_per_1k_output_tokens=0.0015,
    ),
    'gpt-3.5-turbo-0125': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0005,
        cost_per_1k_output_tokens=0.0015,
    ),
    'gpt-3.5-turbo-1106': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.001,
        cost_per_1k_output_tokens=0.002,
    ),
    'gpt-3.5-turbo-0613': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0015,
        cost_per_1k_output_tokens=0.002,
    ),
    'gpt-3.5-turbo-0301': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0015,
        cost_per_1k_output_tokens=0.002,
    ),
    'gpt-3.5-turbo-16k': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.004,
    ),
    'gpt-3.5-turbo-16k-0613': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.004,
    ),
    'gpt-3.5-turbo-16k-0301': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.004,
    ),
    # GPT-3.5 models
    'text-davinci-003': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    'text-davinci-002': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    'code-davinci-002': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    # GPT-3 instruction-tuned models
    'text-curie-001': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    'text-babbage-001': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM,
    ),
    'text-ada-001': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM,
    ),
    'davinci': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM,
    ),
    'curie': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    'babbage': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    'ada': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    # GPT-3 base models
    'babbage-002': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
    'davinci-002': pg.Dict(
        in_service=True,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM
    ),
}


@lf.use_init_args(['model'])
class OpenAI(lf.LanguageModel):
  """OpenAI model."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      'The name of the model to use.',
  ] = 'gpt-3.5-turbo'

  multimodal: Annotated[
      bool,
      'Whether this model has multimodal support.'
  ] = False

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'OPENAI_API_KEY'."
      ),
  ] = None

  organization: Annotated[
      str | None,
      (
          'Organization. If None, the key will be read from environment '
          "variable 'OPENAI_ORGANIZATION'. Based on the value, usages from "
          "these API requests will count against the organization's quota. "
      ),
  ] = None

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('_api_initialized', None)
    if openai is None:
      raise RuntimeError(
          'Please install "langfun[llm-openai]" to use OpenAI models.'
      )

  @functools.cached_property
  def _api_initialized(self):
    api_key = self.api_key or os.environ.get('OPENAI_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `OPENAI_API_KEY` with your OpenAI API key.'
      )
    openai.api_key = api_key
    org = self.organization or os.environ.get('OPENAI_ORGANIZATION', None)
    if org:
      openai.organization = org
    return True

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'OpenAI({self.model})'

  @property
  def max_concurrency(self) -> int:
    rpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('rpm', 0)
    tpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('tpm', 0)
    return self.rate_to_max_concurrency(
        requests_per_min=rpm, tokens_per_min=tpm
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1k_input_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_input_tokens', None
    )
    cost_per_1k_output_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_output_tokens', None
    )
    if cost_per_1k_output_tokens is None or cost_per_1k_input_tokens is None:
      return None
    return (
        cost_per_1k_input_tokens * num_input_tokens
        + cost_per_1k_output_tokens * num_output_tokens
    ) / 1000

  @classmethod
  def dir(cls):
    assert openai is not None
    return openai.Model.list()

  @property
  def is_chat_model(self):
    """Returns True if the model is a chat model."""
    return self.model.startswith(('o1', 'gpt-4', 'gpt-3.5-turbo'))

  def _get_request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    # Reference:
    # https://platform.openai.com/docs/api-reference/completions/create
    # NOTE(daiyip): options.top_k is not applicable.
    args = dict(
        n=options.n,
        stream=False,
        timeout=self.timeout,
        top_logprobs=options.top_logprobs,
    )
    if options.logprobs:
      # Reasoning models (o1 series) does not support `logprobs` by 2024/09/12.
      if self.model.startswith('o1-'):
        raise RuntimeError('`logprobs` is not supported on {self.model!r}.')
      args['logprobs'] = options.logprobs

    # Completion and ChatCompletion uses different parameter name for model.
    args['model' if self.is_chat_model else 'engine'] = self.model

    if options.temperature is not None:
      args['temperature'] = options.temperature
    if options.max_tokens is not None:
      args['max_tokens'] = options.max_tokens
    if options.top_p is not None:
      args['top_p'] = options.top_p
    if options.stop:
      args['stop'] = options.stop
    if options.random_seed is not None:
      args['seed'] = options.random_seed
    return args

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    assert self._api_initialized
    if self.is_chat_model:
      return self._chat_complete_batch(prompts)
    else:
      return self._complete_batch(prompts)

  def _complete_batch(
      self, prompts: list[lf.Message]
  ) -> list[lf.LMSamplingResult]:

    def _open_ai_completion(prompts):
      assert openai is not None
      response = openai.Completion.create(
          prompt=[p.text for p in prompts],
          **self._get_request_args(self.sampling_options),
      )
      # Parse response.
      samples_by_index = collections.defaultdict(list)
      for choice in response.choices:
        samples_by_index[choice.index].append(
            lf.LMSample(choice.text.strip(), score=choice.logprobs or 0.0)
        )

      n = len(samples_by_index)
      estimated_cost = self.estimate_cost(
          num_input_tokens=response.usage.prompt_tokens,
          num_output_tokens=response.usage.completion_tokens,
      )
      usage = lf.LMSamplingUsage(
          prompt_tokens=response.usage.prompt_tokens // n,
          completion_tokens=response.usage.completion_tokens // n,
          total_tokens=response.usage.total_tokens // n,
          estimated_cost=(
              None if estimated_cost is None else (estimated_cost // n)
          )
      )
      return [
          lf.LMSamplingResult(samples_by_index[index], usage=usage)
          for index in sorted(samples_by_index.keys())
      ]

    return self._parallel_execute_with_currency_control(
        _open_ai_completion,
        [prompts],
        retry_on_errors=(
            ServiceUnavailableError,
            RateLimitError,
            APITimeoutError,
        ),
    )[0]

  def _chat_complete_batch(
      self, prompts: list[lf.Message]
  ) -> list[lf.LMSamplingResult]:
    def _content_from_message(message: lf.Message):
      if self.multimodal:
        content = []
        for chunk in message.chunk():
          if isinstance(chunk, str):
            item = dict(type='text', text=chunk)
          elif isinstance(chunk, lf_modalities.Image):
            uri = chunk.uri or chunk.content_uri
            item = dict(type='image_url', image_url=dict(url=uri))
          else:
            raise ValueError(f'Unsupported modality object: {chunk!r}.')
          content.append(item)
      else:
        content = message.text
      return content

    def _open_ai_chat_completion(prompt: lf.Message):
      request_args = self._get_request_args(self.sampling_options)
      # Users could use `metadata_json_schema` to pass additional
      # request arguments.
      json_schema = prompt.metadata.get('json_schema')
      if json_schema is not None:
        if not isinstance(json_schema, dict):
          raise ValueError(
              f'`json_schema` must be a dict, got {json_schema!r}.'
          )
        if 'title' not in json_schema:
          raise ValueError(
              f'The root of `json_schema` must have a `title` field, '
              f'got {json_schema!r}.'
          )
        request_args.update(
            response_format=dict(
                type='json_schema',
                json_schema=dict(
                    schema=json_schema,
                    name=json_schema['title'],
                    strict=True,
                )
            )
        )
        prompt.metadata.formatted_text = (
            prompt.text
            + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
            + pg.to_json_str(request_args['response_format'], json_indent=2)
        )

      # Prepare messages.
      messages = []
      # Users could use `metadata_system_message` to pass system message.
      system_message = prompt.metadata.get('system_message')
      if system_message:
        system_message = lf.SystemMessage.from_value(system_message)
        messages.append(
            dict(role='system', content=_content_from_message(system_message))
        )
      messages.append(dict(role='user', content=_content_from_message(prompt)))

      assert openai is not None
      response = openai.ChatCompletion.create(messages=messages, **request_args)

      samples = []
      for choice in response.choices:
        logprobs = None
        choice_logprobs = getattr(choice, 'logprobs', None)
        if choice_logprobs:
          logprobs = [
              (
                  t.token,
                  t.logprob,
                  [(tt.token, tt.logprob) for tt in t.top_logprobs],
              )
              for t in choice_logprobs.content
          ]
        samples.append(
            lf.LMSample(
                choice.message.content,
                score=0.0,
                logprobs=logprobs,
            )
        )

      return lf.LMSamplingResult(
          samples=samples,
          usage=lf.LMSamplingUsage(
              prompt_tokens=response.usage.prompt_tokens,
              completion_tokens=response.usage.completion_tokens,
              total_tokens=response.usage.total_tokens,
              estimated_cost=self.estimate_cost(
                  num_input_tokens=response.usage.prompt_tokens,
                  num_output_tokens=response.usage.completion_tokens,
              )
          ),
      )

    return self._parallel_execute_with_currency_control(
        _open_ai_chat_completion,
        prompts,
        retry_on_errors=(
            ServiceUnavailableError,
            RateLimitError,
            APITimeoutError
        ),
    )


class GptO1Preview(OpenAI):
  """GPT-O1."""
  model = 'o1-preview'


class GptO1Preview_20240912(OpenAI):   # pylint: disable=invalid-name
  """GPT O1."""
  model = 'o1-preview-2024-09-12'


class GptO1Mini(OpenAI):
  """GPT O1-mini."""
  model = 'o1-mini'


class GptO1Mini_20240912(OpenAI):   # pylint: disable=invalid-name
  """GPT O1-mini."""
  model = 'o1-mini-2024-09-12'


class Gpt4(OpenAI):
  """GPT-4."""
  model = 'gpt-4'


class Gpt4Turbo(Gpt4):
  """GPT-4 Turbo with 128K context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo'
  multimodal = True


class Gpt4Turbo_20240409(Gpt4Turbo):  # pylint:disable=invalid-name
  """GPT-4 Turbo with 128K context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo-2024-04-09'
  multimodal = True


class Gpt4TurboPreview(Gpt4):
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo-preview'


class Gpt4TurboPreview_20240125(Gpt4TurboPreview):  # pylint: disable=invalid-name
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-0125-preview'


class Gpt4TurboPreview_20231106(Gpt4TurboPreview):  # pylint: disable=invalid-name
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Apr. 2023."""
  model = 'gpt-4-1106-preview'


class Gpt4VisionPreview(Gpt4):
  """GPT-4 Turbo vision preview. 128k context window. Knowledge to Apr. 2023."""
  model = 'gpt-4-vision-preview'
  multimodal = True


class Gpt4VisionPreview_20231106(Gpt4):  # pylint: disable=invalid-name
  """GPT-4 Turbo vision preview. 128k context window. Knowledge to Apr. 2023."""
  model = 'gpt-4-1106-vision-preview'


class Gpt4_20230613(Gpt4):    # pylint:disable=invalid-name
  """GPT-4 @20230613. 8K context window. Knowledge up to 9-2021."""
  model = 'gpt-4-0613'


class Gpt4_32K(Gpt4):       # pylint:disable=invalid-name
  """Latest GPT-4 with 32K context window."""
  model = 'gpt-4-32k'


class Gpt4_32K_20230613(Gpt4_32K):    # pylint:disable=invalid-name
  """GPT-4 @20230613. 32K context window. Knowledge up to 9-2021."""
  model = 'gpt-4-32k-0613'


class Gpt4oMini(OpenAI):
  """GPT-4o Mini."""
  model = 'gpt-4o-mini'
  multimodal = True


class Gpt4oMini_20240718(OpenAI):  # pylint:disable=invalid-name
  """GPT-4o Mini."""
  model = 'gpt-4o-mini-2024-07-18'
  multimodal = True


class Gpt4o(OpenAI):
  """GPT-4o."""
  model = 'gpt-4o'
  multimodal = True


class Gpt4o_20241120(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-11-20."""
  model = 'gpt-4o-2024-11-20'
  multimodal = True


class Gpt4o_20240806(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-08-06."""
  model = 'gpt-4o-2024-08-06'
  multimodal = True


class Gpt4o_20240513(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-05-13."""
  model = 'gpt-4o-2024-05-13'
  multimodal = True


class Gpt35(OpenAI):
  """GPT-3.5. 4K max tokens, trained up on data up to Sep, 2021."""
  model = 'text-davinci-003'


class Gpt35Turbo(Gpt35):
  """Most capable GPT-3.5 model, 10x cheaper than GPT35 (text-davinci-003)."""
  model = 'gpt-3.5-turbo'


class Gpt35Turbo_20240125(Gpt35Turbo):   # pylint:disable=invalid-name
  """GPT-3.5 Turbo @20240125. 16K context window. Knowledge up to 09/2021."""
  model = 'gpt-3.5-turbo-0125'


class Gpt35Turbo_20231106(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gpt3.5 Turbo @20231106. 16K context window. Knowledge up to 09/2021."""
  model = 'gpt-3.5-turbo-1106'


class Gpt35Turbo_20230613(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gpt3.5 Turbo snapshot at 2023/06/13, with 4K context window size."""
  model = 'gpt-3.5-turbo-0613'


class Gpt35Turbo16K(Gpt35Turbo):
  """Latest GPT-3.5 model with 16K context window size."""
  model = 'gpt-3.5-turbo-16k'


class Gpt35Turbo16K_20230613(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gtp 3.5 Turbo 16K 0613."""
  model = 'gpt-3.5-turbo-16k-0613'


class Gpt3(OpenAI):
  """Most capable GPT-3 model (Davinci) 2K context window size.

  All GPT3 models have 2K max tokens and trained on data up to Oct 2019.
  """
  model = 'davinci'


class Gpt3Curie(Gpt3):
  """Very capable, but faster and lower cost than Davici."""
  model = 'curie'


class Gpt3Babbage(Gpt3):
  """Capable of straightforward tasks, very fast and low cost."""
  model = 'babbage'


class Gpt3Ada(Gpt3):
  """Capable of very simple tasks, the fastest/lowest cost among GPT3 models."""
  model = 'ada'



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Python code execution."""

import inspect
import time
import unittest
from langfun.core.coding.python import errors
from langfun.core.coding.python import execution
from langfun.core.coding.python import permissions
import pyglove as pg


class EvaluateTest(unittest.TestCase):

  def test_with_context(self):
    with execution.context(x=1, y=0):
      with execution.context(x=2, z=2):
        self.assertEqual(
            execution.evaluate(
                """
                p = x + y + z
                """,
                # Override value from the context.
                global_vars=dict(z=3),
                outputs_intermediate=True,
            ),
            dict(p=2 + 0 + 3, __result__=2 + 0 + 3, __stdout__=''),
        )

  def test_basics(self):
    self.assertEqual(
        execution.evaluate(
            """
            x = 1
            y = x + 1
            print(y)
            z = x + y
            """,
            outputs_intermediate=True,
        ),
        dict(x=1, y=2, z=3, __result__=3, __stdout__='2\n'),
    )
    self.assertEqual(
        execution.evaluate(
            """
            x = 1
            y = x + 1
            print(y)
            z = x + y
            """,
        ),
        3,
    )

  def test_class_def(self):
    ret = execution.evaluate(
        """
        class A(pg.Object):
          x: int
          y: int
          def __call__(self):
            return self.x + self.y
        """,
        permission=permissions.CodePermission.ALL,
        global_vars=dict(pg=pg),
        outputs_intermediate=True,
    )
    self.assertEqual(list(ret.keys()), ['A', '__result__', '__stdout__'])
    self.assertTrue(issubclass(ret['A'], pg.Object))
    self.assertIs(ret['__result__'], ret['A'])
    self.assertEqual(ret['__stdout__'], '')

  def test_function_def(self):
    ret = execution.evaluate(
        """
        def foo(x, y):
          return x + y

        def bar(z):
          return z + foo(z, z)
        """,
        permission=permissions.CodePermission.ALL,
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['foo', 'bar', '__result__', '__stdout__']
    )
    self.assertTrue(inspect.isfunction(ret['foo']))
    self.assertTrue(inspect.isfunction(ret['bar']))
    self.assertIs(ret['__result__'], ret['bar'])

  def test_function_def_and_call(self):
    ret = execution.evaluate(
        """
        def foo(x, y):
          return x + y

        def bar(z):
          return z + foo(z, z)

        bar(1)
        """,
        permission=permissions.CodePermission.ALL,
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['foo', 'bar', '__result__', '__stdout__']
    )
    self.assertEqual(ret['__result__'], 3)

  def test_complex(self):
    ret = execution.evaluate(
        """
        class A(pg.Object):
          x: int
          y: int
          def __call__(self, z):
            return self.x + self.y + z

        def foo(x, y):
          return x + y
        k = A(1, 2)
        k(foo(3, 4))
        """,
        permission=permissions.CodePermission.ALL,
        global_vars=dict(pg=pg),
        outputs_intermediate=True,
    )
    self.assertEqual(
        list(ret.keys()), ['A', 'foo', 'k', '__result__', '__stdout__']
    )
    self.assertTrue(issubclass(ret['A'], pg.Object))
    self.assertTrue(inspect.isfunction(ret['foo']))
    self.assertIsInstance(ret['k'], pg.Object)
    self.assertEqual(ret['__result__'], 10)

  def test_run_with_error(self):
    with self.assertRaisesRegex(
        errors.CodeError, 'NameError: name .* is not defined'
    ):
      execution.evaluate(
          """
          x = 1
          y = x + z
          """,
          permission=permissions.CodePermission.ALL,
      )
    with self.assertRaisesRegex(errors.CodeError, 'ValueError'):
      execution.evaluate(
          'raise ValueError()', permission=permissions.CodePermission.ALL
      )


class Foo(pg.Object):
  x: int
  y: int


class SandboxCallTest(unittest.TestCase):

  def test_basics(self):
    def f(x, y):
      return x + y
    self.assertEqual(execution.sandbox_call(f, 1, y=2), 3)

  def test_complex_type(self):
    def f(x, y):
      return Foo(x, y)

    self.assertEqual(execution.sandbox_call(f, 1, 2), Foo(1, 2))

  def test_timeout(self):
    def f(x):
      time.sleep(x)

    self.assertIsNone(execution.sandbox_call(f, 0, timeout=1))
    with self.assertRaises(TimeoutError):
      execution.sandbox_call(f, 2, timeout=1)

  def test_raise(self):
    def f(x):
      if x == 0:
        raise ValueError()

    self.assertIsNone(execution.sandbox_call(f, 1))
    with self.assertRaises(ValueError):
      execution.sandbox_call(f, 0)


class CallTest(unittest.TestCase):

  def test_call_without_sandboxing(self):
    def foo(x, y):
      return x + y

    self.assertEqual(
        execution.call(foo, 1, y=2, sandbox=False),
        3
    )

  def test_call_with_sandboxing(self):
    def foo(x, y):
      return x + y

    self.assertEqual(
        execution.call(foo, 1, y=2, sandbox=True),
        3
    )

    def make_cls():
      class A(pg.Object):
        x: str
      return A

    with self.assertRaises(errors.SerializationError):
      execution.call(make_cls, sandbox=True)

  def test_call_with_automatic_sandboxing(self):
    def foo(x, y):
      return x + y

    self.assertEqual(
        execution.call(foo, 1, y=2),
        3
    )

    def make_cls():
      class A(pg.Object):
        x: str
      return A

    self.assertTrue(inspect.isclass(execution.call(make_cls)))


class RunTest(unittest.TestCase):

  def test_run_without_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
            sandbox=False,
        ),
        3,
    )

  def test_run_with_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
            sandbox=True,
        ),
        3,
    )

  def test_run_with_automatic_sandboxing(self):
    self.assertEqual(
        execution.run(
            'x + y',
            global_vars=dict(x=1, y=2),
        ),
        3,
    )

    r = execution.run(
        inspect.cleandoc("""
            def foo(x, y):
              return x + y

            class A(pg.Object):
              x: str
            """),
        global_vars=dict(pg=pg),
        outputs_intermediate=True,
    )
    self.assertTrue(inspect.isfunction(r['foo']))
    self.assertTrue(inspect.isclass(r['A']))


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Groq models."""

import os
from typing import Any
import unittest
from unittest import mock
from langfun.core import modalities as lf_modalities
from langfun.core.llms import groq
import pyglove as pg
import requests


def mock_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs

  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'choices': [{
          'message': {
              'content': [{
                  'type': 'text',
                  'text': (
                      f'hello with temperature={json.get("temperature")}, '
                      f'top_p={json.get("top_p")}, '
                      f'max_tokens={json.get("max_tokens")}, '
                      f'stop={json.get("stop")}.'
                  ),
              }],
          }
      }],
      'usage': {
          'prompt_tokens': 2,
          'completion_tokens': 1,
          'total_tokens': 3,
      },
  }).encode()
  return response


def mock_mm_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs
  v = json['messages'][0]['content'][0]
  image = lf_modalities.Image.from_uri(v['image_url'])

  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'choices': [
          {
              'message': {
                  'content': [{
                      'type': 'text',
                      'text': image.uri,
                  }],
              }
          }
      ],
      'usage': {
          'prompt_tokens': 2,
          'completion_tokens': 1,
          'total_tokens': 3,
      },
  }).encode()
  return response


def mock_requests_post_error(status_code, error_type, error_message):
  def _mock_requests(url: str, json: dict[str, Any], **kwargs):
    del url, json, kwargs
    response = requests.Response()
    response.status_code = status_code
    response._content = pg.to_json_str(
        {
            'error': {
                'type': error_type,
                'message': error_message,
            }
        }
    ).encode()
    return response

  return _mock_requests


class AuthropicTest(unittest.TestCase):

  def test_basics(self):
    self.assertEqual(groq.GroqMistral_8x7B().model_id, 'mixtral-8x7b-32768')
    self.assertEqual(groq.GroqMistral_8x7B().max_concurrency, 16)

  def test_api_key(self):
    lm = groq.GroqMistral_8x7B()
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      lm('hi')

    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post

      lm = groq.GroqMistral_8x7B(api_key='fake key')
      self.assertRegex(lm('hi').text, 'hello.*')

      os.environ['GROQ_API_KEY'] = 'abc'
      lm = groq.GroqMistral_8x7B()
      self.assertRegex(lm('hi').text, 'hello.*')
      del os.environ['GROQ_API_KEY']

  def test_call(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post
      lm = groq.GroqLlama3_70B(api_key='fake_key')
      response = lm(
          'hello',
          temperature=0.0,
          max_tokens=1024,
          top_k=0.1,
          top_p=0.2,
          stop=['\n'],
      )
      self.assertEqual(
          response.text,
          (
              'hello with temperature=0.0, top_p=0.2, '
              "max_tokens=1024, stop=['\\n']."
          ),
      )
      self.assertIsNotNone(response.usage)
      self.assertIsNotNone(response.usage.prompt_tokens, 2)
      self.assertIsNotNone(response.usage.completion_tokens, 1)
      self.assertIsNotNone(response.usage.total_tokens, 3)

  def test_mm_call(self):
    with mock.patch('requests.Session.post') as mock_mm_request:
      mock_mm_request.side_effect = mock_mm_requests_post
      lm = groq.GroqLlama3_70B(multimodal=True, api_key='fake_key')
      response = lm(lf_modalities.Image.from_uri('https://fake/image.jpg'))
      self.assertEqual(response.text, 'https://fake/image.jpg')

  def test_call_errors(self):
    for status_code, error_type, error_message in [
        (429, 'rate_limit', 'Rate limit exceeded.'),
        (503, 'service_unavailable', 'Service unavailable.'),
        (500, 'bad_request', 'Bad request.'),
    ]:
      with mock.patch('requests.Session.post') as mock_mm_request:
        mock_mm_request.side_effect = mock_requests_post_error(
            status_code, error_type, error_message
        )
        lm = groq.GroqLlama3_70B(api_key='fake_key')
        with self.assertRaisesRegex(
            Exception, f'{status_code}:.*{error_type}'
        ):
          lm('hello', max_attempts=1)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
import unittest

from langfun.core.eval.v2 import checkpointing
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import runners as runners_lib  # pylint: disable=unused-import
from langfun.core.eval.v2 import test_helper
import pyglove as pg

Example = example_lib.Example


class StateWriterTest(unittest.TestCase):

  def test_basic(self):
    file = os.path.join(tempfile.gettempdir(), 'test.jsonl')
    writer = checkpointing.StateWriter(file)
    example = Example(id=1, input=pg.Dict(x=1), output=2)
    writer.add(example)
    del writer
    self.assertTrue(pg.io.path_exists(file))

  def test_error_handling(self):
    file = os.path.join(tempfile.gettempdir(), 'test_error_handling.jsonl')
    writer = checkpointing.StateWriter(file)
    writer.add(Example(id=1, input=pg.Dict(x=1), output=2))

    def f():
      raise ValueError('Intentional error')

    try:
      writer.add(f())
    except ValueError:
      del writer

    self.assertTrue(pg.io.path_exists(file))
    with pg.io.open_sequence(file, 'r') as f:
      self.assertEqual(len(list(iter(f))), 1)


class CheckpointingTest(unittest.TestCase):

  def test_checkpointing(self):
    root_dir = os.path.join(tempfile.gettempdir(), 'test_checkpointing')
    experiment = test_helper.test_experiment()
    checkpoint_filename = 'checkpoint.jsonl'
    checkpointer = checkpointing.Checkpointer(checkpoint_filename)
    run = experiment.run(
        root_dir, 'new', runner='sequential', plugins=[checkpointer]
    )
    self.assertEqual(len(checkpointer._state_writer), 0)
    num_processed = {}
    for leaf in experiment.leaf_nodes:
      ckpt = run.output_path_for(leaf, checkpoint_filename)
      self.assertTrue(pg.io.path_exists(ckpt))
      with pg.io.open_sequence(ckpt) as f:
        self.assertEqual(
            len(list(iter(f))),
            leaf.progress.num_completed - leaf.progress.num_failed
        )
      if leaf.id not in num_processed:
        self.assertEqual(leaf.progress.num_skipped, 0)
        num_processed[leaf.id] = leaf.progress.num_processed

    # Run again, should skip existing.
    _ = experiment.run(
        root_dir, 'latest', runner='sequential', plugins=[checkpointer]
    )
    self.assertEqual(len(checkpointer._state_writer), 0)
    for leaf in experiment.leaf_nodes:
      self.assertEqual(leaf.progress.num_skipped, num_processed[leaf.id])


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun Component."""

import contextlib
import dataclasses
import threading
from typing import Annotated, Any, ContextManager, Iterator, Type
import pyglove as pg


# Default value marker that indicates to raise error.
RAISE_IF_HAS_ERROR = (pg.MISSING_VALUE,)


class Component(pg.Object):
  """Base class for langfun components."""

  # Override __repr__ format to use inferred values when available.
  __repr_format_kwargs__ = dict(
      compact=True,
      use_inferred=True,
  )

  # Override __str__ format to use inferred values when available.
  __str_format_kwargs__ = dict(
      compact=False,
      verbose=False,
      use_inferred=True,
  )

  # Allow symbolic assignment, which invalidates the object and recomputes
  # states upon update.
  allow_symbolic_assignment = True

  # Class property that indicates whether to use `sym_eq` for `__eq__`,
  # `sym_ne` for `__ne__`, and `sym_hash` for `__hash__`.
  use_symbolic_comparison = False

  def __init_subclass__(cls):
    super().__init_subclass__()

    # Find class attributes that do not have annotations but are `Component`,
    # and treat them as symbolic fields.
    additional_fields = []
    for attr_name in dir(cls):
      if (
          attr_name.startswith('_')
          or attr_name.isupper()
          or attr_name in cls.__schema__.fields
      ):
        continue
      attr_value = getattr(cls, attr_name)
      if isinstance(attr_value, pg.Inferentiable):
        value_spec = pg.typing.Any()
      elif isinstance(attr_value, Component):
        value_spec = pg.typing.Object(Component)
      else:
        value_spec = None
      if value_spec:
        field = pg.typing.create_field((attr_name, value_spec))
        field.value.set_default(attr_value)
        additional_fields.append(field)
    if additional_fields:
      cls.update_schema(additional_fields)

  def _on_bound(self):
    super()._on_bound()
    self._tls = threading.local()

  def _sym_inferred(self, key: str, **kwargs):
    """Override to allow attribute to access scoped value.

    Args:
      key: attribute name.
      **kwargs: Optional keyword arguments for value inference.

    Returns:
      The value of the symbolic attribute. If not available, returns the
        default value.

    Raises:
      AttributeError: If the attribute does not exist or contextual attribute
        is not ready.
    """
    if key not in self._sym_attributes:
      raise AttributeError(key)

    # Step 1: Try use value from `self.override`.
    # The reason is that `self.override` is short-lived and explicitly specified
    # by the user in scenarios like `LangFunc.render`, which should not be
    # affected by `lf.context`.
    v = _get_scoped_value(self._tls, _CONTEXT_OVERRIDES, key)
    if v is not None:
      return v.value

    # Step 2: Try use value from `lf.context` with `override_attrs`.
    # This gives users a chance to override the bound attributes of components
    # from the top, allowing change of bindings without modifying the code
    # that produces the components.
    override = get_contextual_override(key)
    if override and override.override_attrs:
      return override.value

    # Step 3: Try use value from the symbolic tree, starting from self to
    # the root of the tree.
    # Step 4: If the value is not present, use the value from `context()` (
    # override_attrs=False).
    # Step 5: Otherwise use the default value from `ContextualAttribute`.
    return super()._sym_inferred(key, context_override=override, **kwargs)

  def override(
      self, **kwargs) -> ContextManager[dict[str, 'ContextualOverride']]:
    """Context manager to override the attributes of this component."""
    vs = {k: ContextualOverride(v) for k, v in kwargs.items()}
    return _contextual_scope(self._tls, _CONTEXT_OVERRIDES, **vs)

  def __getattribute__(self, name: str) -> Any:
    """Override __getattribute__ to deal with class attribute override."""
    if not name.startswith('_') and hasattr(self.__class__, name):
      tls = self.__dict__.get('_tls', None)
      if tls is not None:
        v = _get_scoped_value(tls, _CONTEXT_OVERRIDES, name)
        if v is not None:
          return v.value
    return super().__getattribute__(name)


_global_tls = threading.local()

_CONTEXT_OVERRIDES = 'context_overrides'


@dataclasses.dataclass(frozen=True)
class ContextualOverride:
  """Value marker for contextual override for an attribute."""

  # Overridden value.
  value: Any

  # If True, this override will apply to both current scope and nested scope,
  # meaning current `lf.context` will take precedence over all nested
  # `lf.context` on this attribute.
  cascade: bool = False

  # If True, this override will apply to attributes that already have values.
  override_attrs: bool = False


def context(
    *,
    cascade: bool = False,
    override_attrs: bool = False,
    **variables,
) -> ContextManager[dict[str, ContextualOverride]]:
  """Context manager to provide overriden values for contextual attributes.

  Args:
    cascade: If True, this override will apply to both current scope and nested
      scope, meaning that this `lf.context` will take precedence over all
      nested `lf.context` on the overriden variables.
    override_attrs: If True, this override will apply to attributes that already
      have values. Otherwise overridden variables will only be used for
      contextual attributes whose values are not present.
    **variables: Key/values as override for contextual attributes.

  Returns:
    A dict of attribute names to their contextual overrides.
  """
  vs = {}
  for k, v in variables.items():
    if not isinstance(v, ContextualOverride):
      v = ContextualOverride(v, cascade, override_attrs)
    vs[k] = v
  return _contextual_scope(_global_tls, _CONTEXT_OVERRIDES, **vs)


def use_settings(
    *,
    cascade: bool = False,
    **settings,
) -> ContextManager[dict[str, ContextualOverride]]:
  """Shortcut method for overriding component attributes.

  Args:
    cascade: If True, this override will apply to both current scope and nested
      scope, meaning that this `lf.context` will take precedence over all
      nested `lf.context` on the overriden variables.
    **settings: Key/values as override for component attributes.

  Returns:
    A dict of attribute names to their contextual overrides.
  """
  return context(cascade=cascade, override_attrs=True, **settings)


def get_contextual_override(var_name: str) -> ContextualOverride | None:
  """Returns the overriden contextual value in current scope."""
  return _get_scoped_value(_global_tls, _CONTEXT_OVERRIDES, var_name)


def context_value(var_name: str, default: Any = RAISE_IF_HAS_ERROR) -> Any:
  """Returns the value of a variable defined in `lf.context`."""
  override = get_contextual_override(var_name)
  if override is None:
    if default == RAISE_IF_HAS_ERROR:
      raise KeyError(f'{var_name!r} does not exist in current context.')
    return default
  return override.value


def all_contextual_values() -> dict[str, Any]:
  """Returns all contextual values provided from `lf.context` in scope."""
  overrides = getattr(_global_tls, _CONTEXT_OVERRIDES, {})
  return {k: v.value for k, v in overrides.items()}


@contextlib.contextmanager
def _contextual_scope(
    tls: threading.local, tls_key, **variables
) -> Iterator[dict[str, ContextualOverride]]:
  """Context manager to set variables within a scope."""
  previous_values = getattr(tls, tls_key, {})
  current_values = dict(previous_values)
  for k, v in variables.items():
    old_v = current_values.get(k, None)
    if old_v and old_v.cascade:
      v = old_v
    current_values[k] = v
  try:
    setattr(tls, tls_key, current_values)
    yield current_values
  finally:
    setattr(tls, tls_key, previous_values)


def _get_scoped_value(
    tls: threading.local, tls_key: str, var_name: str, default: Any = None
) -> ContextualOverride:
  """Gets the value for requested variable from current scope."""
  scoped_values = getattr(tls, tls_key, {})
  return scoped_values.get(var_name, default)


class ContextualAttribute(
    pg.symbolic.ValueFromParentChain, pg.views.HtmlTreeView.Extension
):
  """Attributes whose values are inferred from the context of the component.

  Please see go/langfun-component#attribute-value-retrieval for details.
  """

  NO_DEFAULT = (pg.MISSING_VALUE,)

  type: Annotated[Type[Any] | None, 'An optional type constraint.'] = None

  default: Any = NO_DEFAULT

  def value_from(
      self,
      parent,
      *,
      context_override: ContextualOverride | None = None,
      **kwargs,
  ):
    if parent not in (None, self.sym_parent) and isinstance(parent, Component):
      # Apply original search logic along the component containing chain.
      return super().value_from(parent, **kwargs)
    elif parent is None:
      # When there is no value inferred from the symbolic tree.
      # Search context override, and then attribute-level default.
      if context_override:
        return context_override.value
      if self.default == ContextualAttribute.NO_DEFAULT:
        return pg.MISSING_VALUE
      return self.default
    else:
      return pg.MISSING_VALUE

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      parent: Any = None,
      root_path: pg.KeyPath | None = None,
      **kwargs,
  ) -> pg.Html:
    inferred_value = pg.MISSING_VALUE
    if isinstance(parent, pg.Symbolic) and root_path:
      inferred_value = parent.sym_inferred(root_path.key, pg.MISSING_VALUE)

    if inferred_value is not pg.MISSING_VALUE:
      kwargs.pop('name', None)
      return view.render(
          inferred_value, parent=self,
          root_path=pg.KeyPath('<inferred>', root_path),
          **view.get_passthrough_kwargs(**kwargs)
      )
    return pg.Html.element(
        'div',
        [
            '(not available)',
        ],
        css_classes=['unavailable-contextual'],
    )

  def _html_tree_view_config(self) -> dict[str, Any]:
    return pg.views.HtmlTreeView.get_kwargs(
        super()._html_tree_view_config(),
        dict(
            collapse_level=1,
        )
    )

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .contextual-attribute {
          color: purple;
        }
        .unavailable-contextual {
          color: gray;
          font-style: italic;
        }
        """
    ]


# NOTE(daiyip): Returning Any instead of `lf.ContextualAttribute` to avoid
# pytype check error as `contextual()` can be assigned to any type.
def contextual(
    type: Type[Any] | None = None,  # pylint: disable=redefined-builtin
    default: Any = ContextualAttribute.NO_DEFAULT,
) -> Any:
  """Value marker for a contextual attribute."""
  return ContextualAttribute(type=type, default=default, allow_partial=True)


# Decorator for setting the positional arguments for Component.
use_init_args = pg.use_init_args



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun features on Python code parsing and execution."""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member

from langfun.core.coding.python.errors import CodeError

from langfun.core.coding.python.permissions import CodePermission
from langfun.core.coding.python.permissions import permission
from langfun.core.coding.python.permissions import get_permission

from langfun.core.coding.python.parsing import PythonCodeParser

from langfun.core.coding.python.execution import context
from langfun.core.coding.python.execution import get_context
from langfun.core.coding.python.execution import evaluate
from langfun.core.coding.python.execution import sandbox_call
from langfun.core.coding.python.execution import call
from langfun.core.coding.python.execution import run

from langfun.core.coding.python.generation import PythonCode
from langfun.core.coding.python.generation import PythonFunction

from langfun.core.coding.python.correction import correct
from langfun.core.coding.python.correction import run_with_correction
from langfun.core.coding.python.correction import CodeWithError

# pylint: enable=g-importing-member
# pylint: enable=g-bad-import-order



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for language models through REST APIs."""

import functools
from typing import Annotated, Any, Callable

import langfun.core as lf
import requests


class REST(lf.LanguageModel):
  """REST-based language model."""

  api_endpoint: Annotated[
      str,
      'The endpoint of the REST API.'
  ]

  request: Annotated[
      Callable[[lf.Message, lf.LMSamplingOptions], dict[str, Any]],
      'A function to convert a Langfun message to a JSON request.'
  ]

  result: Annotated[
      Callable[[dict[str, Any]], lf.LMSamplingResult],
      'A function to convert a JSON response to an LMSamplingResult.'
  ]

  model: Annotated[
      str | None,
      'Model ID.'
  ] = None

  headers: Annotated[
      dict[str, Any] | None,
      'The headers for the REST API.'
  ] = None

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return self.model or 'unknown'

  @functools.cached_property
  def _api_initialized(self) -> bool:
    """Returns whether the API is initialized."""
    self._initialize()
    return True

  def _initialize(self) -> None:
    """Initializes the API. Subclasses can override."""

  @functools.cached_property
  def _session(self) -> requests.Session:
    assert self._api_initialized
    s = requests.Session()
    s.headers.update(self.headers or {})
    return s

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('_session', None)
    self.__dict__.pop('_api_initialized', None)

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    assert self._api_initialized
    return self._parallel_execute_with_currency_control(
        self._sample_single, prompts
    )

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    try:
      response = self._session.post(
          self.api_endpoint,
          json=self.request(prompt, self.sampling_options),
          timeout=self.timeout,
      )
      return self._parse_response(response)
    except ConnectionError as e:
      raise lf.LMError(str(e)) from e

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if status_code == 429:
      error_cls = lf.RateLimitError
    elif status_code in (
        500,  # Server side issue (might be bug).
        502,  # Bad gateway (upstream issue, might retry).
        503,  # Servers currently under load, retry after a brief wait.
        529,  # Overloaded, retry after a brief wait.
    ):
      error_cls = lf.TemporaryLMError
    else:
      error_cls = lf.LMError
    return error_cls(f'{status_code}: {content}')

  def _parse_response(self, response: requests.Response) -> lf.LMSamplingResult:
    """Parses Anthropic's response."""
    if response.status_code == 200:
      return self.result(response.json())
    else:
      raise self._error(response.status_code, response.content)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gemini models exposed through Google Generative AI APIs."""

import abc
import functools
import os
from typing import Annotated, Any, Literal

import langfun.core as lf
from langfun.core import modalities as lf_modalities
import pyglove as pg


try:
  import google.generativeai as genai   # pylint: disable=g-import-not-at-top
  BlobDict = genai.types.BlobDict
  GenerativeModel = genai.GenerativeModel
  Completion = getattr(genai.types, 'Completion', Any)
  ChatResponse = getattr(genai.types, 'ChatResponse', Any)
  GenerateContentResponse = getattr(genai.types, 'GenerateContentResponse', Any)
  GenerationConfig = genai.GenerationConfig
except ImportError:
  genai = None
  BlobDict = Any
  GenerativeModel = Any
  Completion = Any
  ChatResponse = Any
  GenerationConfig = Any
  GenerateContentResponse = Any


@lf.use_init_args(['model'])
class GenAI(lf.LanguageModel):
  """Language models provided by Google GenAI."""

  model: Annotated[
      Literal[
          'gemini-pro',
          'gemini-pro-vision',
          'text-bison-001',
          'chat-bison-001',
          'gemini-1.5-pro-latest',
          'gemini-1.5-flash-latest',
          'gemini-exp-1114',
      ],
      'Model name.',
  ]

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'GOOGLE_API_KEY'. "
          'Get an API key at https://ai.google.dev/tutorials/setup'
      ),
  ] = None

  supported_modalities: Annotated[
      list[str],
      'A list of MIME types for supported modalities'
  ] = []

  # Set the default max concurrency to 8 workers.
  max_concurrency = 8

  def _on_bound(self):
    super()._on_bound()
    if genai is None:
      raise RuntimeError(
          'Please install "langfun[llm-google-genai]" to use '
          'Google Generative AI models.'
      )
    self.__dict__.pop('_api_initialized', None)

  @functools.cached_property
  def _api_initialized(self):
    assert genai is not None
    api_key = self.api_key or os.environ.get('GOOGLE_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `GOOGLE_API_KEY` with your Google Cloud API key. '
          'Check out '
          'https://cloud.google.com/api-keys/docs/create-manage-api-keys '
          'for more details.'
      )
    genai.configure(api_key=api_key)
    return True

  @classmethod
  def dir(cls) -> list[str]:
    """Lists generative models."""
    assert genai is not None
    return [
        m.name.lstrip('models/')
        for m in genai.list_models()
        if (
            'generateContent' in m.supported_generation_methods
            or 'generateText' in m.supported_generation_methods
            or 'generateMessage' in m.supported_generation_methods
        )
    ]

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return self.model

  @property
  def resource_id(self) -> str:
    """Returns a string to identify the resource for rate control."""
    return self.model_id

  def _generation_config(self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Creates generation config from langfun sampling options."""
    return GenerationConfig(
        candidate_count=options.n,
        temperature=options.temperature,
        top_p=options.top_p,
        top_k=options.top_k,
        max_output_tokens=options.max_tokens,
        stop_sequences=options.stop,
    )

  def _content_from_message(
      self, prompt: lf.Message
  ) -> list[str | BlobDict]:
    """Gets Evergreen formatted content from langfun message."""
    formatted = lf.UserMessage(prompt.text)
    formatted.source = prompt

    chunks = []
    for lf_chunk in formatted.chunk():
      if isinstance(lf_chunk, str):
        chunks.append(lf_chunk)
      elif isinstance(lf_chunk, lf_modalities.Mime):
        try:
          modalities = lf_chunk.make_compatible(
              self.supported_modalities + ['text/plain']
          )
          if isinstance(modalities, lf_modalities.Mime):
            modalities = [modalities]
          for modality in modalities:
            if modality.is_text:
              chunk = modality.to_text()
            else:
              chunk = BlobDict(
                  data=modality.to_bytes(),
                  mime_type=modality.mime_type
              )
            chunks.append(chunk)
        except lf.ModalityError as e:
          raise lf.ModalityError(f'Unsupported modality: {lf_chunk!r}') from e
      else:
        raise lf.ModalityError(f'Unsupported modality: {lf_chunk!r}')
    return chunks

  def _response_to_result(
      self, response: GenerateContentResponse | pg.Dict
  ) -> lf.LMSamplingResult:
    """Parses generative response into message."""
    samples = []
    for candidate in response.candidates:
      chunks = []
      for part in candidate.content.parts:
        # TODO(daiyip): support multi-modal parts when they are available via
        # Gemini API.
        if hasattr(part, 'text'):
          chunks.append(part.text)
      samples.append(lf.LMSample(lf.AIMessage.from_chunks(chunks), score=0.0))
    return lf.LMSamplingResult(samples)

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    assert self._api_initialized, 'Vertex AI API is not initialized.'
    return self._parallel_execute_with_currency_control(
        self._sample_single,
        prompts,
    )

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    """Samples a single prompt."""
    model = _GOOGLE_GENAI_MODEL_HUB.get(self.model)
    input_content = self._content_from_message(prompt)
    response = model.generate_content(
        input_content,
        generation_config=self._generation_config(self.sampling_options),
    )
    return self._response_to_result(response)


class _LegacyGenerativeModel(pg.Object):
  """Base for legacy GenAI generative model."""

  model: str

  def generate_content(
      self,
      input_content: list[str | BlobDict],
      generation_config: GenerationConfig,
  ) -> pg.Dict:
    """Generate content."""
    segments = []
    for s in input_content:
      if not isinstance(s, str):
        raise ValueError(f'Unsupported modality: {s!r}')
      segments.append(s)
    return self.generate(' '.join(segments), generation_config)

  @abc.abstractmethod
  def generate(
      self, prompt: str, generation_config: GenerationConfig) -> pg.Dict:
    """Generate response based on prompt."""


class _LegacyCompletionModel(_LegacyGenerativeModel):
  """Legacy GenAI completion model."""

  def generate(
      self, prompt: str, generation_config: GenerationConfig
  ) -> pg.Dict:
    assert genai is not None
    completion: Completion = genai.generate_text(
        model=f'models/{self.model}',
        prompt=prompt,
        temperature=generation_config.temperature,
        top_k=generation_config.top_k,
        top_p=generation_config.top_p,
        candidate_count=generation_config.candidate_count,
        max_output_tokens=generation_config.max_output_tokens,
        stop_sequences=generation_config.stop_sequences,
    )
    return pg.Dict(
        candidates=[
            pg.Dict(content=pg.Dict(parts=[pg.Dict(text=c['output'])]))
            for c in completion.candidates
        ]
    )


class _LegacyChatModel(_LegacyGenerativeModel):
  """Legacy GenAI chat model."""

  def generate(
      self, prompt: str, generation_config: GenerationConfig
  ) -> pg.Dict:
    assert genai is not None
    response: ChatResponse = genai.chat(
        model=f'models/{self.model}',
        messages=prompt,
        temperature=generation_config.temperature,
        top_k=generation_config.top_k,
        top_p=generation_config.top_p,
        candidate_count=generation_config.candidate_count,
    )
    return pg.Dict(
        candidates=[
            pg.Dict(content=pg.Dict(parts=[pg.Dict(text=c['content'])]))
            for c in response.candidates
        ]
    )


class _ModelHub:
  """Google Generative AI model hub."""

  def __init__(self):
    self._model_cache = {}

  def get(
      self, model_name: str
  ) -> GenerativeModel | _LegacyGenerativeModel:
    """Gets a generative model by model id."""
    assert genai is not None
    model = self._model_cache.get(model_name, None)
    if model is None:
      model_info = genai.get_model(f'models/{model_name}')
      if 'generateContent' in model_info.supported_generation_methods:
        model = genai.GenerativeModel(model_name)
      elif 'generateText' in model_info.supported_generation_methods:
        model = _LegacyCompletionModel(model_name)
      elif 'generateMessage' in model_info.supported_generation_methods:
        model = _LegacyChatModel(model_name)
      else:
        raise ValueError(f'Unsupported model: {model_name!r}')
      self._model_cache[model_name] = model
    return model


_GOOGLE_GENAI_MODEL_HUB = _ModelHub()


#
# Public Gemini models.
#


_IMAGE_TYPES = [
    'image/png',
    'image/jpeg',
    'image/webp',
    'image/heic',
    'image/heif',
]

_AUDIO_TYPES = [
    'audio/aac',
    'audio/flac',
    'audio/mp3',
    'audio/m4a',
    'audio/mpeg',
    'audio/mpga',
    'audio/mp4',
    'audio/opus',
    'audio/pcm',
    'audio/wav',
    'audio/webm'
]

_VIDEO_TYPES = [
    'video/mov',
    'video/mpeg',
    'video/mpegps',
    'video/mpg',
    'video/mp4',
    'video/webm',
    'video/wmv',
    'video/x-flv',
    'video/3gpp',
]

_PDF = [
    'application/pdf',
]


class GeminiExp_20241114(GenAI):  # pylint: disable=invalid-name
  """Gemini Experimental model launched on 11/14/2024."""

  model = 'gemini-exp-1114'
  supported_modalities = _PDF + _IMAGE_TYPES + _AUDIO_TYPES + _VIDEO_TYPES


class GeminiPro1_5(GenAI):  # pylint: disable=invalid-name
  """Gemini Pro latest model."""

  model = 'gemini-1.5-pro-latest'
  supported_modalities = _PDF + _IMAGE_TYPES + _AUDIO_TYPES + _VIDEO_TYPES


class GeminiFlash1_5(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash latest model."""

  model = 'gemini-1.5-flash-latest'
  supported_modalities = _PDF + _IMAGE_TYPES + _AUDIO_TYPES + _VIDEO_TYPES


class GeminiPro(GenAI):
  """Gemini Pro model."""

  model = 'gemini-pro'


class GeminiProVision(GenAI):
  """Gemini Pro vision model."""

  model = 'gemini-pro-vision'
  supported_modalities = _IMAGE_TYPES + _VIDEO_TYPES


class Palm2(GenAI):
  """PaLM2 model."""

  model = 'text-bison-001'


class Palm2_IT(GenAI):  # pylint: disable=invalid-name
  """PaLM2 instruction-tuned model."""

  model = 'chat-bison-001'



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for llama cpp models."""

import typing
import unittest
from unittest import mock

from langfun.core.llms import llama_cpp


def mock_requests_post(url: str, json: typing.Dict[str, typing.Any], **kwargs):
  del kwargs

  class TEMP:
    @property
    def status_code(self):
      return 200

    def json(self):
      return {"content": json["prompt"] + "\n" + url}

  return TEMP()


class LlamaCppRemoteTest(unittest.TestCase):
  """Tests for the LlamaCppRemote model."""

  def test_call_completion(self):
    with mock.patch("requests.Session.post") as mock_request:
      mock_request.side_effect = mock_requests_post
      lm = llama_cpp.LlamaCppRemote("http://127.0.0.1:8080")
      [result] = lm.sample(["hello"], n=2)
      self.assertEqual(
          len(result.samples),
          2
      )
      self.assertEqual(
          str(result.samples[0].response),
          "hello\nhttp://127.0.0.1:8080/completion",
      )

  def test_model_id(self):
    lm = llama_cpp.LlamaCppRemote("http://127.0.0.1:8080")
    self.assertEqual(lm.model_id, "LLaMAC++()")
    lm = llama_cpp.LlamaCppRemote("xxx", model="x")
    self.assertEqual(lm.model_id, "LLaMAC++(x)")


if __name__ == "__main__":
  unittest.main()
